from nd2reader import ND2Reader
import bioformats
import javabridge
import tifffile
import numpy as np
from pathlib import Path

import multiprocessing as mp

def get_editable_omexml(path):
    """
    Parse OMEXML header data from a Bio-Formats-compatible file.

    Used to parse metadata from .nd2 file and pass on to .ome.tiff file.
    """
   
    o = bioformats.get_omexml_metadata(path)
    new_omexml = bioformats.OMEXML(o)

    return new_omexml

def convert_nd2_to_ome_tiff(path, outpath):
    """
    Convert an nd2 file to an .ome.tif by saving each tile under a separate series.

    Uses the BigTIFF format to write the new .ome.tiff file.
    """

    path_to_nd2 = str(path)

    omexml = get_editable_omexml(path_to_nd2)
    limit = omexml.image_count
    
    xml_dict = tifffile.xml2dict(omexml.to_xml())
    with ND2Reader(path_to_nd2) as images:
        print(images.sizes)
        images.iter_axes = 'v'
        images.bundle_axes = "zcyx"
        with tifffile.TiffWriter(outpath, bigtiff=True) as tif:
            for series, tile in enumerate(images):
                # TODO: seems like, for some weird reason, the tiles are rotated incorrectly? Why?
                # Maybe it has to do with the order of the bundle axes

                rotated_tile = np.rot90(tile, k=1, axes=(2, 3))
                series_metadata = xml_dict["OME"]["Image"][series]
                tif.save(rotated_tile.astype(np.uint16), contiguous=False, metadata=series_metadata)

def slice_ome_tiff(input_filepath, output_filepath, start, end, use_bigtiff=False):
    """
    Build a new .ome.tiff with only the series that fall in the range start:end included (+ metadata).
    """
    # TODO: Add an IndexError or whatever later

    input_ome_tiff = tifffile.TiffFile(input_filepath)

    xml = input_ome_tiff.ome_metadata
    xml_dict = tifffile.xml2dict(xml)

    print(input_filepath)
    with tifffile.TiffWriter(output_filepath, bigtiff=use_bigtiff) as tif:
        for index in range(start, end):
            print(index)
            tile = input_ome_tiff.series[index].asarray()
            tile_metadata = xml_dict["OME"]["Image"][index]
            tif.save(tile.astype(np.uint16), contiguous=False, metadata=tile_metadata)

def convert_nd2_all_samples(config, parallelize=0):
    """
    """
    # TODO: Fill in docstring

    workspace_directory = config["workspace_directory"]
    input_directory = Path(config["data_directory"])
    output_directory = Path(workspace_directory, "unstitched")
    output_directory.mkdir(exist_ok=True)

    samples = config["samples"]

    # TODO: Figure out if it's possible to parallelize by sample here.
    # TODO: Figure out better error handling here (e.g. catch error and
    # kill JVM if error occurs)

    javabridge.start_vm(class_path=bioformats.JARS)
   
    processes = [] 
    for sample in samples:
        if parallelize > 0:
            process = mp.Process(target=convert_nd2_single_sample, args=(sample, input_directory, output_directory, parallelize - 1))
            process.start() 
            processes.append(process)
 
        else:
            convert_nd2_single_sample(sample, input_directory, output_directory)

    for process in processes:
        process.join()

    javabridge.kill_vm()

def convert_nd2_single_sample(sample, input_directory, output_directory, parallelize=0):
    """
    """

    rounds = sample["rounds"]
    sample_name = sample["name"]

    sample_output_directory = output_directory / sample_name
    sample_output_directory.mkdir(exist_ok=True)

    processes = []
    for round_index, imaging_round in enumerate(rounds):
        filename = imaging_round["filename"]

        input_filepath = (input_directory / filename).with_suffix(".nd2")
        output_filepath = (sample_output_directory / filename).with_suffix(".ome.tif")

        if parallelize > 0:
            process = mp.Process(target=convert_nd2_to_ome_tiff, args=(input_filepath, output_filepath))
            process.start() 
            processes.append(process)
        else:
            convert_nd2_to_ome_tiff(input_filepath, output_filepath)

    for process in processes:
        process.join()

def slice_ome_tiff_all_samples(config, start, end):
    """
    """
    # TODO: Fill in docstring

    workspace_directory = config["workspace_directory"]
    input_directory = Path(workspace_directory, "unstitched")
    output_directory = Path(config["slice_directory"])
    output_directory.mkdir(exist_ok=True)

    samples = config["samples"]

    # TODO: Figure out if it's possible to parallelize by sample here.
    
    for sample in samples:
        slice_ome_tiff_single_sample(sample, input_directory, output_directory, start, end)

def slice_ome_tiff_single_sample(sample, input_directory, output_directory, start, end):
    rounds = sample["rounds"]
    sample_name = sample["name"]

    sample_input_directory = input_directory / sample_name

    for round_index, imaging_round in enumerate(rounds):
        filename = imaging_round["filename"]

        input_filepath = (sample_input_directory / filename).with_suffix(".ome.tif")
        output_filepath = (output_directory / filename).with_suffix(".ome.tif")

        slice_ome_tiff(input_filepath, output_filepath, start, end, use_bigtiff=True)



import numpy as np
import imageio
from pathlib import Path
import multiprocessing as mp

from starfish import data, FieldOfView
from starfish.types import Axes, Features
from starfish.image import Filter
from starfish.core.imagestack.imagestack import ImageStack
from starfish.spots import DecodeSpots, FindSpots
from starfish import Codebook

def find_spots_all_samples(config, parallelize=0):
    """
    """
    # TODO: Fill in docstring

    workspace_directory = config["workspace_directory"]
    input_directory = Path(workspace_directory, "stitched")
    output_directory = Path(workspace_directory, "spots_only")
    output_directory.mkdir(exist_ok=True)

    samples = config["samples"]

    # TODO: Figure out if it's possible to parallelize by sample here.
    if parallelize > 0:
        num_processes = mp.cpu_count()
        print(num_processes)
        processes = []
        for sample in samples:
            process = mp.Process(target=find_spots_single_sample, args=(sample, input_directory, output_directory, parallelize - 1))
            process.start()
            processes.append(process)

        for process in processes:
            process.join()
    else:
        for sample in samples:
            find_spots_single_sample(sample, input_directory, output_directory)

def find_spots_single_sample(sample, input_directory, output_directory, parallelize=0):
    """
    """
    # TODO: Fill in docstring

    sample_name = sample["name"]
    rounds = sample["rounds"]

    processes = []
    for round_index, imaging_round in enumerate(rounds, start=1):
        if parallelize > 0:
            process = mp.Process(target=find_spots_in_round, args=(sample_name, round_index, imaging_round, input_directory, output_directory))
            process.start()
            processes.append(process)
        else:
            find_spots_in_round(sample_name, round_index, imaging_round, input_directory, output_directory)

    for process in processes:
        process.join()

def find_spots_in_round(sample_name, round_index, imaging_round, input_directory, output_directory):
    """
    """

    round_directory = ("round%d" % round_index)
    sample_input_subdirectory = input_directory / sample_name / round_directory
    sample_output_subdirectory = output_directory / sample_name / round_directory
    sample_output_subdirectory.mkdir(parents=True, exist_ok=True)

    channels = imaging_round["channels"]
    filename = imaging_round["filename"]
    reference_channel = imaging_round["reference_channel"] 
    for channel_index, channel in enumerate(channels):
        input_filename = filename + ("_fused_tp_0_ch_%d.tif" % channel_index)
        output_filename = channel + ".tif"
        input_path = sample_input_subdirectory / input_filename
        output_path = sample_output_subdirectory / output_filename

        if channel_index == reference_channel:
            image_stack = imageio.volread(input_path)
            max_filtered_image_stack = image_stack.max(0)
            imageio.imsave(output_path, max_filtered_image_stack)
        else:
            find_spots(input_path, output_path)


def find_spots(input_path, output_path, intensity_percentile=99.995, filter_width=2, small_peak_min=4, small_peak_max=100,
               big_peak_min=25, big_peak_max=10000, small_peak_dist=2, big_peak_dist=0.75, block_dim_fraction=0.25,
               spot_pad_pixels=2, keep_existing=False):
    """
    Find and keep only spots from stitched images.

    """

    image_stack = imageio.volread(input_path)

    print(image_stack.shape)
    thresholded_image = np.copy(image_stack)
    
    _, height, width = image_stack.shape

    threshold = np.percentile(thresholded_image, intensity_percentile)
    thresholded_image[thresholded_image > threshold] = threshold + (np.log(thresholded_image[thresholded_image > threshold] - threshold)/np.log(1.1)).astype(thresholded_image.dtype)

    #May need to fiddle with the sigma parameters in each step, depending on the image.
    
    #High Pass Filter (Background Subtraction)
    gaussian_high_pass = Filter.GaussianHighPass(sigma=(1, filter_width, filter_width), is_volume=True)
    
    # enhance brightness of spots
    laplace_filter = Filter.Laplace(sigma=(0.2, 0.5, 0.5), is_volume=True)
    local_max_peakfinder_small = FindSpots.LocalMaxPeakFinder(
        min_distance=small_peak_dist,
        stringency=0,
        min_obj_area=small_peak_min,
        max_obj_area=small_peak_max,
        min_num_spots_detected=2500,
        is_volume=True,
        verbose=True
    )

    local_max_peakfinder_big = FindSpots.LocalMaxPeakFinder(
        min_distance=big_peak_dist,
        stringency=0,
        min_obj_area=big_peak_min,
        max_obj_area=big_peak_max,
        min_num_spots_detected=2500,
        is_volume=True,
        verbose=True
    )

    synthetic_codebook= Codebook.synthetic_one_hot_codebook(n_round=1, n_channel=1, n_codes=1)
    decoder = DecodeSpots.PerRoundMaxChannel(codebook=synthetic_codebook)

    block_dimension = int(max(thresholded_image.shape) * block_dim_fraction)
    spot_coordinates= np.zeros((0, 2), dtype=np.int64)

    # Finding spots by block_dimension x block_dimension size blocks
    # We skip the blocks at the edges with the - 1 (TODO: pad to full block size)
    for row in range(0, height - 1, block_dimension):
        for column in range(0, width - 1, block_dimension):
            # Cutout block and expand dimensions for channel and round
            block = thresholded_image[np.newaxis, np.newaxis, :, row:row+block_dimension, column:column+block_dimension]
            images = ImageStack.from_numpy(block)
            high_pass_filtered = gaussian_high_pass.run(images, verbose=False, in_place=False)
            laplace = laplace_filter.run(high_pass_filtered, in_place=False,verbose=False)

            small_spots = local_max_peakfinder_small.run(laplace.reduce({Axes.ZPLANE}, func="max"))
            decoded_intensities = decoder.run(spots=small_spots)
            small_spot_coords = np.stack([decoded_intensities[Axes.Y.value], decoded_intensities[Axes.X.value]]).T
            
            big_spots = local_max_peakfinder_big.run(laplace.reduce({Axes.ZPLANE}, func="max"))
            decoded_intensities = decoder.run(spots=big_spots)
            big_spot_coords = np.stack([decoded_intensities[Axes.Y.value], decoded_intensities[Axes.X.value]]).T
            
            all_spot_coords = np.vstack([small_spot_coords, big_spot_coords])
            all_spot_coords += (row, column)

            spot_coordinates = np.vstack([spot_coordinates, all_spot_coords])

    # Copying over only non-zero pixels
    image_spots = np.zeros((height, width), dtype=np.uint16)
    for spot_coordinate in spot_coordinates:
        spot_column, spot_row = spot_coordinate
        for row in range(max(0, spot_column-spot_pad_pixels), min(spot_column+spot_pad_pixels+1, height)):
            for column in range(max(0, spot_row-spot_pad_pixels), min(spot_row+spot_pad_pixels+1, width)):
                # Max projecting over z-stack
                image_spots[row, column] = image_stack[:, row, column].max(0)

    imageio.imsave(output_path, image_spots)

    return image_spots



from util import *

def register_and_tile_images(tissue_directory_regex, tile_size, blank_round_number):
    round_subdirectory_regex = "round*/"
    blank_round_string = "round%d/" % blank_round_number
    corrected_filename = "flat_field_corrected.tiff"
    registered_filename = "registered.tiff"

    tissue_directories = glob.glob(tissue_directory_regex)

    for tissue_directory in tissue_directories:
        Images = []
        #print(tissue_directory)
        round_directory_regex = tissue_directory + round_subdirectory_regex
        blank_round_directory = os.path.join(tissue_directory, blank_round_string)
        round_directories = [directory for directory in glob.glob(round_directory_regex) if directory != blank_round_directory]

        for round_directory in round_directories:
            round_image_filepath = os.path.join(round_directory, corrected_filename)
            round_image = imageio.imread(round_image_filepath)
            Images.append(round_image)
        
        min_shape = (min(im.shape[0] for im in Images), min(im.shape[1] for im in Images))
        Images = np.array([im[:min_shape[0],:min_shape[1]] for im in Images])
       
        # TODO: How to calculate dapi index? It's zero for now
        dapi_slices = Images[:, :, :, 0]
        first_dapi = dapi_slices[0]
        shifts = find_all_shifts(first_dapi, dapi_slices)
        print("Shifts calculated.")
        ImagesAligned = np.array([apply_shifts(Images[index], shifts[index]) for index in range(len(Images))])
        ImagesAligned = crop_all_shifted(ImagesAligned, shifts)

        for round_index, round_directory in enumerate(round_directories):
            print(round_directory)
            round_output_filepath = os.path.join(round_directory, registered_filename)
            round_output_image = ImagesAligned[round_index]
            imageio.imwrite(round_output_filepath, ImagesAligned[round_index])
            #max_value = np.iinfo(Images[0].dtype).max
            #round_output_image = ormalize_image_scale(round_output_image, max_value)

            os.makedirs(os.path.join(round_directory, "tiles"), exist_ok=True)
            tiles, new_fov_pattern = tilify(ImagesAligned[round_index], tile_size)
            for tile, new_fov in zip(tiles, new_fov_pattern.flatten()):
                array_filepath = os.path.join(round_directory, "tiles", "fov_%d.npy" % new_fov)
                np.save(array_filepath, tile)
                image_filepath = os.path.join(round_directory, "tiles", "fov_%d.tiff" % new_fov)
                imageio.imwrite(image_filepath, tile)
            np.save(os.path.join(round_directory, "tiles", "modified_fov_pattern.npy"), new_fov_pattern)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--tile-size', help='Size of output images',type=int,default=1024)
    parser.add_argument('--tissue-directory-regex', help='Path to directory with subdirs for parsed images in each tissue')
    parser.add_argument('--blank-round-number', help='Path to directory with subdirs for parsed images in each tissue')
    parser.set_defaults(save_tiles=False)
    parser.set_defaults(save_stitched=False)
    args, _ = parser.parse_known_args()

    register_and_tile_images(args.tissue_directory_regex, args.tile_size, args.blank_round+number)



import numpy as np
import os
import argparse
import glob
from ast import literal_eval
import imageio
from scipy.ndimage import fourier_shift
from scipy import ndimage

def filter(im, filter_size):
	if filter_size > 0:
		if len(im.shape) == 3:
			for i in range(im.shape[2]):
				im[:,:,i] = ndimage.median_filter(im[:,:,i], filter_size)
	return im

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--basepath', help='Path to directory with subdirs for parsed images in each tissue')
	parser.add_argument('--channels', help='Comma separated ordered list of channels, e.g. Conf-488,conf-561,conf-640,conf-405')
	parser.add_argument('--filter-size', help='Size of median filter',type=int,default=8)
	parser.add_argument('--jobnum', help='Job index for running on cluster',type=int,default=-1)
	parser.add_argument('--blank-round', help='Round in which blank images were acquired (will use bigger filter on these)')
	parser.add_argument('--keep-existing', help='Keep existing stitched files (if present)',dest='keep_existing', action='store_true')
	parser.set_defaults(keep_existing=False)
	args,_ = parser.parse_known_args()
	channels = args.channels.split(',')
	FP = glob.glob(os.path.join(args.basepath,'*','*','DAPI.conf-405','TileConfiguration.registered.txt'))
	if args.jobnum > -1:
		FP = FP[args.jobnum:args.jobnum+1]
	for fp in FP:
		outfile = '/'.join(fp.split('/')[:-2] + ['stitched.tiff'])
		if (args.keep_existing and not os.path.isfile(outfile)) or (not args.keep_existing):
			fovs = []
			coords = []
			f = open(fp)
			header = [f.readline() for _ in range(4)]
			for line in f:
				ls = line.strip().split(';')
				fovs.append(ls[0])
				coords.append(literal_eval(ls[2].strip()))
			f.close()
			coords = np.rint(coords).astype(np.int)
			coords[:,0] -= coords[:,0].min()
			coords[:,1] -= coords[:,1].min()
			fov_size = imageio.imread(fp.replace('TileConfiguration.registered.txt',fovs[0])).shape
			stitched_size = (coords.max(0) + fov_size)[::-1]
			stitched_size = np.hstack([stitched_size,len(channels)])
			stitched = np.zeros(stitched_size,dtype=np.uint16)
			for c,channel in enumerate(channels):
				for fov,coord in zip(fovs,coords):
					fp_channel_fov = glob.glob(os.path.join('/'.join(fp.split('/')[:-2] + ['*.%s' % channel,fov])))
					if len(fp_channel_fov) == 1:
						stitched[coord[1]:coord[1]+fov_size[1],coord[0]:coord[0]+fov_size[0],c] = imageio.imread(fp_channel_fov[0])
			if fp.split('/')[-3] == args.blank_round:
				stitched = filter(stitched, args.filter_size*2)
			else:
				stitched = filter(stitched, args.filter_size)
			imageio.imwrite(outfile, stitched)
			print(outfile)


import numpy as np
import os
import argparse
import glob
import imageio
from scipy.sparse import csr_matrix, save_npz


if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--basepath', help='Path to directory with subdirs for parsed images in each tissue')
	parser.add_argument('--tissues', help='Comma-separated list of tissue numbers to include')
	parser.add_argument('--mask-file', help='Mask file (or prefix), eg ExpandedNuclei_* or DAPI.tiff', default='All_Composite.tiff')
	args,_ = parser.parse_known_args()
	for t in args.tissues.split(','):
		tissue = 'tissue%s' % t
		if '*' in args.mask_file:
			# separate file for each cell
			data = []
			row_ind = []
			col_ind = []
			FP = glob.glob(os.path.join(args.basepath,tissue,'segmented',args.mask_file))
			for fp in FP:
				cell_num = int(fp.split('_')[-1].split('.')[0])-1
				im = imageio.imread(fp).flatten()
				mask = np.where(im)[0]
				data.append(np.ones(len(mask),dtype=np.float32))
				row_ind.append(np.ones(len(mask),dtype=np.int)*cell_num)
				col_ind.append(mask)
			data = np.hstack(data)
			row_ind = np.hstack(row_ind)
			col_ind = np.hstack(col_ind)
		else:
			im = imageio.imread('%s/%s/segmented/%s' % (args.basepath,tissue,args.mask_file)).flatten()
			col_ind = np.where(im > im.min())[0]
			data = np.ones(len(col_ind),dtype=np.float32)
			if im.dtype not in (np.dtype('float32'), np.dtype('float64')):
				row_ind = im[col_ind].astype(np.int)-1
			else:
				# values may have overflowed to floats, but are still unique
				row_ind = []
				mask_dict = {}
				for i in im[col_ind]:
					if i not in mask_dict:
						mask_dict[i] = len(mask_dict)
					row_ind.append(mask_dict[i])
				row_ind = np.array(row_ind)
		X = csr_matrix((data,(row_ind,col_ind)),shape=[len(np.unique(row_ind)),len(im)],dtype=np.float32)
		save_npz('%s/%s/segmented/cell_masks.npz' % (args.basepath,tissue),X)
		print(tissue,X.shape,len(data))





#import spot_finder_starfish as sfs
#sfs.find_spots("BigStitcherResults/B3_round1_fused_tp_0_ch_1.tif", "FilteredCompositeImages/B3_round1_fused_tp_0_ch_1.tif")

import imageio
import numpy as np

from starfish.types import Axes, Features
from starfish.core.imagestack.imagestack import ImageStack
from starfish.spots import DecodeSpots, FindSpots

def local_max_peakfinder_test(path, intensity_percentile=99.995, filter_width=2, small_peak_min=4, small_peak_max=100,
               big_peak_min=25, big_peak_max=10000, small_peak_dist=2, big_peak_dist=0.75, block_dim_fraction=0.5,
               spot_pad_pixels=2, keep_existing=False):
    """
    Debugging FindSpots.LocalMaxPeakFinder
    """

    image_stack = imageio.volread(path)
    
    # enhance brightness of spots
    local_max_peakfinder_small = FindSpots.LocalMaxPeakFinder(
        min_distance=small_peak_dist,
        stringency=0,
        min_obj_area=small_peak_min,
        max_obj_area=small_peak_max,
        min_num_spots_detected=2500,
        is_volume=False,
        verbose=False
    )

    images = ImageStack.from_numpy(image_stack[np.newaxis, np.newaxis, :])
    premature_spots = local_max_peakfinder_small.run(images.reduce({Axes.ZPLANE}, func="max"))

if __name__ == "__main__":
    local_max_peakfinder_test("BigStitcherResults/B3_round1_fused_tp_0_ch_1.tif")



import numpy as np
import os
import argparse
import glob
from collections import defaultdict
from skimage.feature import register_translation
from skimage.transform import rescale
from scipy.ndimage import fourier_shift
import imageio
from ast import literal_eval

def find_all_shifts(ref_image,other_images):
	shifts = []
	for im2 in other_images:
		shifts.append(find_shift(ref_image,im2))
	return shifts

def find_shift(im1,im2,scale=0.25):
	im1_r = rescale(im1,scale)
	im2_r = rescale(im2,scale)
	shift, error, diffphase = register_translation(im1_r, im2_r)
	return shift/scale

def apply_shifts(im,shift):
	offset_image = []
	for i in range(im.shape[2]):
		offset_im = fourier_shift(np.fft.fftn(im[:,:,i]), shift)
		offset_image.append(np.fft.ifftn(offset_im))
	offset_image = np.transpose(offset_image,axes=[1,2,0]).astype(np.float32)
	return offset_image

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--basepath', help='Path to directory with subdirs for parsed images in each tissue')
	parser.add_argument('--tissues', help='Comma-separated list of tissue numbers to include')
	parser.add_argument('--blank-round', help='Round in which blank images were collected')
	parser.add_argument('--dapi-index', help='Channel index for DAPI images',type=int,default=3)
	parser.add_argument('--multiplier', help='Factors to multiply blanks before subtraction (by channel)',default='1.5,1.15,1.35')
	parser.set_defaults(save_tiles=False)
	parser.set_defaults(save_stitched=False)
	args,_ = parser.parse_known_args()
	factors = np.array(list(literal_eval(args.multiplier)) + [1])
	for t in args.tissues.split(','):
		tissue = 'tissue%s' % t
		FP_blanks = glob.glob(os.path.join('%s/%s/%s/stitched.tiff' % (args.basepath,tissue,args.blank_round)))
		Blanks = [imageio.imread(fp) for fp in FP_blanks]
		FP = glob.glob(os.path.join('%s/%s/round*/stitched.tiff' % (args.basepath,tissue)))
		FP = [fp for fp in FP if fp not in FP_blanks]
		Images = [imageio.imread(fp) for fp in FP]
		min_shape = (min(im.shape[0] for im in Images+Blanks), min(im.shape[1] for im in Images+Blanks))
		Images = [im[:min_shape[0],:min_shape[1]] for im in Images]
		Blanks = [im[:min_shape[0],:min_shape[1]] for im in Blanks]
		Images1 = [im[:,:,args.dapi_index] for im in Images]
		Blanks1 = [im[:,:,args.dapi_index] for im in Blanks]
		shifts = [find_shift(im,Blanks1[0]) for im in Images1]
		max_value = np.iinfo(Images[0].dtype).max
		Blanks = Blanks[0]
		# setting the 1% brightest spots to the max, so that they will certainly be set to zero
		thresh = np.percentile(Blanks,99,axis=0)
		Blanks[Blanks > thresh] = max_value
		for i,im in enumerate(Images):
			shifted_blanks = apply_shifts(Blanks,shifts[i])
			im_new = im - shifted_blanks*factors
			im_new[im_new < 0] = 0
			im_new = np.rint(im_new).astype(im.dtype)
			im_new[:,:,args.dapi_index] = im[:,:,args.dapi_index]
			imageio.imwrite(FP[i].replace('.tiff','.background_subtract.tiff'),im_new)
		print(tissue)



from util import *

def parse_and_z_project(background_path_regex, source_path_regex, fov_order_map_path, context_to_channel_map_path, outpath):
    with open(context_to_channel_map_path, 'r') as handle:
        context_to_channel_map = json.load(handle)
    
    MAX_INTENSITY = np.iinfo(np.uint16).max
   

    background_filepaths = sorted(glob.glob(background_path_regex), key = lambda filepath: filepath.split('/')[-1])
    source_filepaths = sorted(glob.glob(source_path_regex), key = lambda filepath: filepath.split('/')[-1])
   
    fov_order_map = np.load(fov_order_map_path)

    data_for_filepath = {}

    # TODO: Why does it have to be sorted?
    for filepath in (source_filepaths + background_filepaths)[40:]:
        filename = os.path.splitext(filepath)[0]
        relative_filename = filename[filename.rfind('/')+1:].split()
        tissue = ''.join(relative_filename[1:3])
        round_number = ''.join(relative_filename[3:5])
        
        print("Processing nd2 file %s" % filepath)
        file_data = parse_nd2_file(filepath)

        source_images, coordinate_pairs, fields_of_view, channels, microns_per_pixel = [file_data[key] for key in ['aggregated_images', 'coordinate_pairs', 'fields_of_view', 'channels', 'microns_per_pixel']]
        num_fov, num_channels, num_rows, num_columns = source_images.shape

        for channel_index, channel in enumerate(channels):
            for fov_index, fov in enumerate(fields_of_view):
                source_image = source_images[fov_index, channel_index]

                #reduced_source_image = normalize_and_convert_to_16_bit(source_image)
                reduced_source_image = source_image

                if channel == 'conf-405':
                    channel_label = 'DAPI_%s' % round_number
                    context_to_channel_map[tissue][round_number][channel] = channel_label
                else:
                    channel_label = context_to_channel_map[tissue][round_number][channel]
                
                with open(context_to_channel_map_path, 'w') as handle:
                    json.dump(context_to_channel_map, handle, sort_keys=True, indent=4)

                os.makedirs(outpath + '/%s/images/%s/%s.%s' % (tissue, round_number, channel_label, channel), exist_ok=True)
                imageio.imwrite(outpath + '/%s/images/%s/%s.%s/fov_%d.png' % (tissue, round_number, channel_label, channel, fov_index), reduced_source_image)
                os.makedirs(outpath + '/%s/arrays/%s/%s.%s' % (tissue, round_number, channel_label, channel), exist_ok=True)
                np.save(outpath + '/%s/arrays/%s/%s.%s/fov_%d.png' % (tissue, round_number, channel_label, channel, fov_index), reduced_source_image)


if __name__=="__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--background-path-regex', help='Paths to files containing background .nd2 images')
    parser.add_argument('--source-path-regex', help='Paths to files containing source .nd2 files')
    parser.add_argument('--fov-order-map-path', help='.npy array with relative locations of each field-of-view')
    parser.add_argument('--context-to-channel-map-path', help='Path to .pickle file representing map from round to field-of-view')
    parser.add_argument('--outpath', help='Output path')
    args, _ = parser.parse_known_args()
   
    parse_and_z_project(args.background_path_regex, args.source_path_regex, args.fov_order_map_path, args.context_to_channel_map_path, args.outpath)



import numpy as np
from scipy.spatial.distance import correlation
from pathlib import Path
import pickle

    
def get_direct_gene_labels(rounds):
    """
    """
    direct_genes = []
    for round_index, imaging_round in enumerate(rounds, start=1):
        channels = imaging_round["channels"]
        composite_channel_indices = imaging_round.get("composite_channels", [])

        for channel_index, channel in enumerate(channels):
            if channel_index not in composite_channel_indices:
                direct_genes.append(channel)

    direct_genes.sort()

    return direct_genes

def analyze_decompression_all_samples(config, parallelize=0):
    """
    """
    # TODO: write docstring

    workspace_directory = config["workspace_directory"]
    training_directory = Path(config["training_directory"])
    segmentation_directory = Path(workspace_directory, "segmented")
    decompression_directory = Path(workspace_directory, "decompressed")
    output_directory = Path(workspace_directory, "analysis")
    output_directory.mkdir(exist_ok=True)

    samples = config["samples"]

    # TODO: Figure out if it's possible to parallelize by sample here.
    if parallelize > 0:
        num_processes = mp.cpu_count()
        processes = []
        for sample in samples:
            process = mp.Process(target=analyze_decompression_single_sample, args=(sample, segmentation_directory, decompression_directory, output_directory, training_directory))
            process.start()
            processes.append(process)

        for process in processes:
            process.join()
    else:
        for sample in samples:
            analyze_decompression_single_sample(sample, segmentation_directory, decompression_directory, output_directory, training_directory)

def analyze_decompression_single_sample(sample, segmentation_directory, decompression_directory, output_directory, training_directory):
    """
    """

    sample_name = sample["name"]
    sample_segmentation_subdirectory = segmentation_directory / sample_name
    sample_decompression_subdirectory = decompression_directory / sample_name
    sample_output_subdirectory = output_directory / sample_name
    sample_output_subdirectory.mkdir(exist_ok=True)

    direct_measurements_filepath = sample_segmentation_subdirectory / "direct_measurements.npy"
    direct_measurements = np.load(direct_measurements_filepath)
    decompressed_measurements_filepath = sample_decompression_subdirectory / "decompressed_measurements.npy"
    decompressed_measurements = np.load(decompressed_measurements_filepath)

    all_gene_labels_filepath = training_directory / "gene_labels.npy"
    all_gene_labels = np.load(all_gene_labels_filepath)

    rounds = sample["rounds"]
    direct_gene_labels = get_direct_gene_labels(rounds)

    direct_gene_indices = np.in1d(all_gene_labels, direct_gene_labels)
    decompressed_direct_measurements = decompressed_measurements[direct_gene_indices]
    print(decompressed_measurements.shape)
    print(decompressed_direct_measurements.shape)
    print(direct_measurements.shape)
    overall_correlation = 1 - correlation(direct_measurements.flatten(), decompressed_direct_measurements.flatten())

    
    print(overall_correlation)
    results = {"overall_correlation": overall_correlation }
    with open(sample_output_subdirectory / "results.pkl", "wb") as result_file:
        pickle.dump(results, result_file)



import numpy as np
from scipy import ndimage
import imageio
from nd2reader import ND2Reader
import glob,os
from collections import defaultdict
import argparse

def parse_rounds(fp):
	Rounds2Channels = defaultdict(list)
	f = open(fp,encoding='utf-8')
	header = f.readline()
	for line in f:
		ls = line.strip().split(',')
		Rounds2Channels[(ls[0],ls[1])].append(ls[2])
	f.close()
	return Rounds2Channels

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--inputdir', help='Path to input images')
	parser.add_argument('--images-to-process', help='CSV with nd2 files, tissue, round')
	parser.add_argument('--rounds-to-channels', help='CSV list of tissue, round, composition labels')
	parser.add_argument('--outdir', help='Output path')
	parser.add_argument('--jobnum', help='Job index for running on cluster',type=int,default=-1)
	args,_ = parser.parse_known_args()
	Rounds2Channels = parse_rounds(args.rounds_to_channels)
	f = open(args.images_to_process)
	FP = [line.strip().split(',') for line in f]
	f.close()
	FP = [('%s/%s' % (args.inputdir,fp[0]), fp[1], fp[2]) for fp in FP]
	if args.jobnum > -1:
		FP = FP[args.jobnum:args.jobnum+1]
	_=os.system('mkdir %s/FilteredCompositeImages' % args.outdir)
	for fp in sorted(FP):
		tissue = fp[1]
		round = fp[2]
		_=os.system('mkdir %s/FilteredCompositeImages/%s' % (args.outdir,tissue))
		_=os.system('mkdir %s/FilteredCompositeImages/%s/%s' % (args.outdir,tissue,round))
		with ND2Reader(fp[0]) as images:
			# iterate over fields of view, then channels
			# max over the z-stack
			try:
				images.iter_axes = 'vcz'
			except:
				images.iter_axes = 'cz'
			num_z = len(images.metadata['z_levels'])
			I = []
			i = 0
			for fov_chann_z in images:
				if i%num_z == 0:
					if i > 0:
						fov_chann = np.array(Iz).max(0)
						I.append(fov_chann)
					Iz = []
				Iz.append(fov_chann_z)
				i += 1
		fov_chann = np.array(Iz).max(0)
		I.append(fov_chann)
		FOV = images.metadata['fields_of_view']
		Channels = images.metadata['channels']
		nc = len(Rounds2Channels[(tissue,round)]) + 1
		for i,f in enumerate(FOV):
			for j,c in enumerate(Channels[:nc]):
				im = I[i*len(Channels) + j]
				if c == 'conf-405':
					channel = 'DAPI'
				else:
					channel = Rounds2Channels[(tissue,round)][j]
				channel = '%s.%s' % (channel,c)
				if i == 0:
					_= os.system('mkdir %s/FilteredCompositeImages/%s/%s/%s' % (args.outdir,tissue,round,channel))
				imageio.imwrite('%s/FilteredCompositeImages/%s/%s/%s/fov_%d.png' % (args.outdir,tissue,round,channel,f),im)
		print(tissue,round)




import numpy as np
import os
import argparse
import glob
import imageio

def apply_threshold(im,x0,x1):
	dt = im.dtype
	max_value = np.iinfo(dt).max
	im[im < x0] = x0
	im[im > x1] = x1
	im = im - x0
	im = np.rint(im/(x1 - x0)*max_value).astype(dt)
	return im	

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--basepath', help='Path to directory with subdirs for parsed images in each tissue')
	parser.add_argument('--thresholds-file', help='File containing channel min and max thresholds to apply to all images')
	parser.add_argument('--zero-width', help='Set (width) edge pixels to zero',type=int,default=6)
	parser.add_argument('--keep-existing', help='Keep existing stitched files (if present)',dest='keep_existing', action='store_true')
	parser.set_defaults(keep_existing=False)
	args,_ = parser.parse_known_args()
	f = open(args.thresholds_file)
	thresholds = [line.strip().split() for line in f]
	f.close()
	thresholds = [[int(t[0]),int(t[1])] for t in thresholds]
	FP = glob.glob(os.path.join(args.basepath,'*','*','stitched.background_subtract.tiff'))
	for fp in FP:
		outfile = fp.replace('.background_subtract.tiff','.threshold.tiff')
		if (args.keep_existing and not os.path.isfile(outfile)) or (not args.keep_existing):
			im = imageio.imread(fp)
			im[:args.zero_width,:] = 0
			im[-args.zero_width:,:] = 0
			im[:,:args.zero_width] = 0
			im[:,-args.zero_width:] = 0
			new_im = np.zeros(im.shape,dtype=im.dtype)
			for i in range(im.shape[-1]):
				im_i = apply_threshold(im[:,:,i],thresholds[i][0],thresholds[i][1])
				new_im[:,:,i] = im_i
			imageio.imwrite(outfile,new_im)
			print(fp)


import numpy as np
import tensorflow as tf
import os
import glob
import argparse
from scipy.sparse import load_npz, csr_matrix, diags
import imageio
from scipy import ndimage

def filter(im, filter_size):
	if filter_size > 0:
		if len(im.shape) == 3:
			for i in range(im.shape[2]):
				im[:,:,i] = ndimage.gaussian_filter(im[:,:,i], filter_size)
		else:
			im = ndimage.gaussian_filter(im, filter_size)
	return im

def _bytes_feature(value):
	return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

def _int64_feature(value):
	if isinstance(value,list):
		return tf.train.Feature(int64_list=tf.train.Int64List(value=value))
	else:
		return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))

def get_cell_masks_in_patch(CellMasks,fov_mask,patch_pos,patch_height,patch_width,min_pixels=144):
	patch_mask = np.zeros(fov_mask.shape,dtype=np.int)
	patch_mask[patch_pos[0]:patch_pos[0]+patch_height, patch_pos[1]:patch_pos[1]+patch_width] = 1
	pm = diags(patch_mask.flatten(),0)
	cell_patch_mask = CellMasks.dot(pm)
	# remove cells with few pixels
	cell_patch_mask = cell_patch_mask[cell_patch_mask.getnnz(1) > min_pixels]
	# adjust indices from full image to within fov
	col_idx = cell_patch_mask.indices % fov_mask.shape[1]
	row_idx = (cell_patch_mask.indices - col_idx) / fov_mask.shape[1]
	col_idx = col_idx - patch_pos[1]
	row_idx = row_idx - patch_pos[0]
	flat_idx = (row_idx*patch_width + col_idx).astype(np.int64)
	n_cells = len(cell_patch_mask.indptr)-1
	sp_embedding_mask_row = []
	sp_embedding_mask_col = []
	sp_embedding_mask_values = []
	for i in range(n_cells):
		for j in range(cell_patch_mask.indptr[i],cell_patch_mask.indptr[i+1]):
			sp_embedding_mask_row.append(i)
			sp_embedding_mask_col.append(j-cell_patch_mask.indptr[i])
			sp_embedding_mask_values.append(flat_idx[j])
	if len(sp_embedding_mask_col) > 0:
		max_pixels_per_cell = max(sp_embedding_mask_col) + 1
	else:
		max_pixels_per_cell = 0
	return n_cells, max_pixels_per_cell, sp_embedding_mask_row, sp_embedding_mask_col, sp_embedding_mask_values

def write_record_from_numpy(image,validation,validation_indices,tissue,fov,fov_mask,CellMasks,tfrecords_filename,patch_height=256,patch_width=256):
	tissue_byte = tf.compat.as_bytes(tissue)
	writer = tf.python_io.TFRecordWriter(tfrecords_filename + '.tfrecords')
	fov_pos = (np.where(fov_mask.sum(1))[0].min(), np.where(fov_mask.sum(0))[0].min())
	for i in range(0,image.shape[0],patch_height):
		for j in range(0,image.shape[1],patch_width):
			image_raw = image[i:i+patch_height,j:j+patch_width]
			# Don't write records for patches with nothing in them
			if image_raw.sum() > 0:
				patch_pos = (i + fov_pos[0], j + fov_pos[1])
				sp_embedding = get_cell_masks_in_patch(CellMasks,fov_mask,patch_pos,patch_height,patch_width)
				validation_raw = validation[i:i+patch_height,j:j+patch_width]
				h,w,c = image_raw.shape
				image_raw = image_raw.tostring()
				validation_raw = validation_raw.tostring()
				example = tf.train.Example(features=tf.train.Features(feature={'raw_data': _bytes_feature(image_raw),
				'height': _int64_feature(h),
				'width': _int64_feature(w),
				'channels': _int64_feature(c),
				'tissue': _bytes_feature(tissue_byte),
				'fov': _int64_feature(fov),
				'row_offset': _int64_feature(i),
				'col_offset': _int64_feature(j),
				'raw_validation': _bytes_feature(validation_raw),
				'validation_indices': _int64_feature(validation_indices),
				'n_cells': _int64_feature(sp_embedding[0]),
				'max_pixels_per_cell': _int64_feature(sp_embedding[1]),
				'sp_embedding_mask_row': _int64_feature(sp_embedding[2]),
				'sp_embedding_mask_col': _int64_feature(sp_embedding[3]),
				'sp_embedding_mask_values': _int64_feature(sp_embedding[4])}))
				writer.write(example.SerializeToString())
	writer.close()

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--patch-width', help='Patch width',type=int,default=256)
	parser.add_argument('--patch-height', help='Patch height',type=int,default=256)
	parser.add_argument('--filter-size', help='Size of median filter (0 for no filtering)',type=float,default=1.75)
	parser.add_argument('--basepath', help='Path to filtered composite images')
	parser.add_argument('--tmppath', help='Tmp path for arrays_aligned_filtered',default=None)
	parser.add_argument('--genes', help='Path to gene labels')
	parser.add_argument('--tissues', help='Comma-separated list of tissue numbers to include')
	parser.add_argument('--fovs', help='Comma-separated list of fovs to include',default=None)
	parser.add_argument('--outdir', help='Sub-Directory for writing TFRecords',default='tfrecords')
	parser.add_argument('--binary', dest='binary', action='store_true')
	parser.set_defaults(binary=False)
	args,_ = parser.parse_known_args()
	f = open(args.genes)
	Genes = [line.strip() for line in f]
	f.close()
	if args.tmppath is None:
		array_path = args.basepath
	else:
		array_path = args.tmppath
	for t in args.tissues.split(','):
		tissue = 'tissue%s' % t
		os.system('mkdir %s/%s/%s' % (args.basepath,tissue,args.outdir))
		CellMasks = load_npz('%s/%s/segmented/cell_masks.npz' % (args.basepath,tissue))
		full_size = imageio.imread('%s/%s/stitched_aligned_filtered/DAPI.tiff' % (args.basepath,tissue)).shape
		if args.fovs is None:
			FP = glob.glob(os.path.join('%s/%s/arrays_aligned_filtered/fov_*.Composite_0.*' % (array_path,tissue)))
			fovs = [fp.split('/')[-1].split('.')[0].split('_')[-1] for fp in FP]
		else:
			fovs = [int(s) for s in args.fovs.split(',')]
		fov_pattern = np.load('%s/%s/modified_fov_pattern.npy' % (args.basepath,tissue))
		for fov in fovs:
			FP = glob.glob(os.path.join('%s/%s/arrays_aligned_filtered/fov_%s.*' % (array_path,tissue,fov)))
			if len(FP) > 0:
				FP_composite = [fp for fp in FP if 'Composite' in fp.split('/')[-1]]
				if '.npy' in FP_composite[0]:
					x = np.load(FP_composite[0])
				elif '.tiff' in FP_composite[0]:
					x = imageio.imread(FP_composite[0])
				width,height = x.shape
				Im_FOV = np.zeros((width,height,len(FP_composite)),dtype=np.float32)
				for fp in FP_composite:
					channel_num = int(fp.split('/')[-1].split('.')[1].split('_')[1])
					if '.npy' in fp:
						x = np.load(fp).astype(np.float32)
					elif '.tiff' in fp:
						x = imageio.imread(fp).astype(np.float32)
					if args.binary:
						x = (x > 0).astype(np.float32)
					x = filter(x,args.filter_size)
					Im_FOV[:,:,channel_num] = x
				FP_validation = [fp for fp in FP if ('Composite' not in fp.split('/')[-1]) and ('DAPI' not in fp)]
				FP_validation = [fp for fp in FP_validation if fp.split('/')[-1].split('.')[1] in Genes]
				Im_FOV_validation = np.zeros((width,height,len(FP_validation)),dtype=np.float32)
				indices = []
				for i,fp in enumerate(FP_validation):
					channel_num = Genes.index(fp.split('/')[-1].split('.')[1])
					if '.npy' in fp:
						x = np.load(fp).astype(np.float32)
					elif '.tiff' in fp:
						x = imageio.imread(fp).astype(np.float32)
					if args.binary:
						x = (x > 0).astype(np.float32)
					x = filter(x,args.filter_size)
					Im_FOV_validation[:,:,i] = x
					indices.append(channel_num)
				fov_mask = np.zeros((width*fov_pattern.shape[0],height*fov_pattern.shape[1]),dtype=np.bool)
				fov_idx = np.where(fov_pattern == int(fov))
				fov_mask[fov_idx[0][0]*width:(fov_idx[0][0] + 1)*width, fov_idx[1][0]*height:(fov_idx[1][0] + 1)*height] = True
				fov_mask = fov_mask[:full_size[0],:full_size[1]]
				write_record_from_numpy(Im_FOV,Im_FOV_validation,indices,tissue,int(fov),fov_mask,CellMasks,'%s/%s/%s/fov_%s' % (args.basepath,tissue,args.outdir,fov), patch_width=args.patch_width, patch_height=args.patch_height)
				print(tissue,fov)



import glob,os
import argparse

composite_label_dict = {'ch0': 'Composite_0',
						'ch1': 'Composite_1',
						'ch2': 'Composite_2',
						'ch3': 'Composite_3',
						'ch4': 'Composite_4',
						'ch5': 'Composite_5',
						'ch6': 'Composite_6',
						'ch7': 'Composite_7',
						'ch8': 'Composite_8',
						'ch9': 'Composite_9',
						'ch10': 'Composite_10'
}

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--path', help='Path to directory with fused images for each round / channel')
	parser.add_argument('--extension', help='Image file extension',default='.tif')
	parser.add_argument('--num-channels', help='Number of channels in filename',type=int,default=3)
	parser.add_argument('--out-path', help='Path to save output')
	args,_ = parser.parse_known_args()
	Tissues = {}
	outfile = open(args.out_path,'w')
	FP = glob.glob(os.path.join(args.path,'*'+args.extension))
	for fp in sorted(FP):
		file = fp.split('/')[-1].split('.')[0]
		well = file.split('_')[0]
		round = file.split('_')[1]
		if well not in Tissues:
			Tissues[well] = 'tissue%d' % (len(Tissues) + 1)
		tissue = Tissues[well]
		channels = file.split('_')[2:2+args.num_channels]
		channel_num = int(file.split('_')[-1])-1
		if channel_num > -1:
			channel = channels[channel_num]
			if channel in composite_label_dict:
				channel = composite_label_dict[channel]
		else:
			channel = 'DAPI'
		_=outfile.write(' '.join([fp,tissue,round,channel]) + '\n')
	outfile.close()
		
		


import imagej
import argparse
import glob
import os

if __name__ == "__main__":
	parser = argparse.ArgumentParser()
	parser.add_argument('--source-paths', help='Paths to directories for each round for each tissue')
	parser.add_argument('--cols', help='Number of columns')
	parser.add_argument('--rows', help='Number of rows')
	parser.add_argument('--fijipath', help='Path to Fiji.app')
	parser.add_argument('--grid-type',help='Stitching pattern',default='snake-by-rows')
	parser.add_argument('--overlap',help='Percent overlap in tiles',type=int,default=10)
	args, _ = parser.parse_known_args()
	grid = ' '.join(args.grid_type.split('-'))
	round_directories = glob.glob(args.source_paths)

	# ImageJ/Fiji setup
	ij = imagej.init(args.fijipath)
	IJM_EXTENSION = ".ijm"
	STITCHING_MACRO = """
	#@ String sourceDirectory
	#@ String outputDirectory
	#@ int overlap
	#@ String grid
	#@ int cols
	#@ int rows

	print(outputDirectory);
	function action(sourceDirectory, outputDirectory, genericFilename, filename, coordinateFilename) {
	run("Grid/Collection stitching", "type=[Grid: " + grid + "] order=[Right & Down ] grid_size_x=" + cols + " grid_size_y=" + rows + " tile_overlap=" + overlap + " first_file_index_i=0 directory=" + sourceDirectory + " file_names=" + genericFilename + " output_textfile_name=" + coordinateFilename + " fusion_method=[Linear Blending] regression_threshold=0.30 max/avg_displacement_threshold=2.50 absolute_displacement_threshold=3.50 subpixel_accuracy compute_overlap computation_parameters=[Save computation time (but use more RAM)] image_output=[Fuse and display]");
	saveAs("png", outputDirectory + filename);
	close();
	}

	action(sourceDirectory, outputDirectory, "fov_{i}.png", "/reference_stitch", "TileConfiguration.txt");

	"""
	cols = args.cols
	rows = args.rows
	overlap = args.overlap
	for round_directory in round_directories:
		try:
			dapi_directory = glob.glob(os.path.join(round_directory, "DAPI*/"))[0]
			print(dapi_directory)
			args = {
			'sourceDirectory': dapi_directory,
			'outputDirectory': round_directory,
			'overlap': overlap,
			'grid': grid,
			'cols': cols,
			'rows': rows
			}

			result = ij.py.run_macro(STITCHING_MACRO, args)
		except:
			pass



import numpy as np
from collections import defaultdict
from skimage.registration import phase_cross_correlation
from skimage.transform import rescale
from scipy.ndimage import fourier_shift
import imageio

import multiprocessing as mp
from pathlib import Path

# def parse_rounds(fp):
#     Rounds2Channels = defaultdict(list)
#     f = open(fp,encoding='utf-8')
#     header = f.readline()
#     for line in f:
#         ls = line.strip().split(',')
#         Rounds2Channels[(ls[0],ls[1])].append(ls[2])
#     f.close()
#     return Rounds2Channels

def find_all_shifts(reference_image, other_images):
    """
    """

    shifts = []
    for other_image in other_images:
        shifts.append(find_shift(reference_image, other_image))

    return shifts

def find_shift(reference_image, other_image, scale=0.25):
    """
    """
    # TODO: Fill out docstring. Also why is there a scale parameter here? Is it to save computation?
    
    rescaled_reference_image = rescale(reference_image, scale)
    rescaled_other_image = rescale(other_image, scale)
    shift, error, diffphase = phase_cross_correlation(rescaled_reference_image, rescaled_other_image)
    scaled_shift = shift / scale

    return scaled_shift

def find_and_apply_shifts(ref_image,other_images,dapi_index,block_size=500,block_pad=250):
    im1 = ref_image[:,:,dapi_index]
    ImagesAligned = [ref_image.astype(np.float32)]
    for im2 in other_images:
        im2_shift = np.zeros(im2.shape,dtype=np.float32)
        for i in range(0,im1.shape[0],block_size):
            for j in range(0,im1.shape[1],block_size):
                im1_block = im1[max(0,i-block_pad):i+block_size+block_pad,max(0,j-block_pad):j+block_size+block_pad]
                im2_block = im2[i:i+block_size,j:j+block_size]
                if (im1_block.sum() > 0) and (im2_block.sum() > 0):
                    pad0 = im1_block.shape[0] - im2_block.shape[0]
                    pad1 = im1_block.shape[1] - im2_block.shape[1]
                    pad0 = (int(pad0/2), int(pad0/2 + pad0%2))
                    pad1 = (int(pad1/2), int(pad1/2 + pad1%2))
                    im2_block = np.pad(im2_block, (pad0, pad1, (0,0)), 'constant').astype(im1_block.dtype)
                    shift, error, diffphase = phase_cross_correlation(im1_block, im2_block[:,:,dapi_index])
                    offset_image = apply_shift_to_stack(im2_block,shift)
                    im2_shift[max(0,i-block_pad):i+block_size+block_pad,max(0,j-block_pad):j+block_size+block_pad] += offset_image
        ImagesAligned.append(im2_shift)
    return ImagesAligned

def apply_shift_to_stack(stack, shift):
    shifting_function = lambda layer: np.fft.ifftn(fourier_shift(np.fft.fftn(layer), shift))
    shifted_stack = np.zeros(stack.shape)

    for channel_index in range(stack.shape[-1]):
        shifted_stack[..., channel_index] = shifting_function(stack[..., channel_index])

    return shifted_stack.astype(np.float32)

def apply_shift(image, shift):
    shifted_image = np.fft.ifftn(fourier_shift(np.fft.fftn(image), shift))

    return shifted_image.real.astype(np.uint16)

def get_shift_crop_index(shifts):
    minimum_shifts = np.min(shifts, axis = 0).astype(np.int)
    maximum_shifts = np.max(shifts, axis = 0).astype(np.int)

    print(shifts)
    print(minimum_shifts)
    print(maximum_shifts)

    x_index = slice(None if maximum_shifts[0] <= 0 else maximum_shifts[0], None if minimum_shifts[0] >= 0 else minimum_shifts[0])
    y_index = slice(None if maximum_shifts[1] <= 0 else maximum_shifts[1], None if minimum_shifts[1] >= 0 else minimum_shifts[1])

    # crop_index = tuple(slice(None if maximum_shifts[index] <= 0 else maximum_shifts[index],
    #                          None if minimum_shifts[index] >= 0 else minimum_shifts[index])
    #                          for index in range(shifts.shape[1]))
    #                   ) 
    # if min_s[0] < 0:
    #   images = images[:min_s[0]]
    # if max_s[0] > 0:
    #   images = images[max_s[0]:]
    # if min_s[1] < 0:
    #   images = images[:,:min_s[1]]
    # if max_s[1] > 0:
    #   images = images[:,max_s[1]:]
    crop_index = (x_index, y_index)

    return crop_index

def normalize_image_scale(im,max_value,thresh_each=False):
    if thresh_each:
        for i in range(im.shape[-1]):
            thresh = np.percentile(im,max_value)
            im[:,:,i][im[:,:,i] > thresh] = thresh
            im[:,:,i] = im[:,:,i]/thresh
        return im.astype(np.float32)
    else:
        return (im/max_value).astype(np.float32)

def break_into_tiles(im,tile_size):
    tiles = []
    tile_pattern = []
    ti = 0
    for i in range(0,im.shape[0],tile_size):
        tile_row = []
        for j in range(0,im.shape[1],tile_size):
            x = im[i:i+tile_size, j:j+tile_size]
            x = np.pad(x, [(0,tile_size-x.shape[0]), (0,tile_size-x.shape[1]), (0,0)], 'constant')
            tiles.append(x)
            tile_row.append(ti)
            ti += 1
        if len(tile_row) > 0:
            tile_pattern.append(tile_row)
    return tiles,tile_pattern

# def rescale_intensities(Images,max_value,dapi_index,eps=1e-4, ptile=95):
#   # rescale all the measurements in each channel according to the min,max in that channel
#   Images = np.array(Images)
#   num_channels = Images.shape[-1]
#   I = Images.reshape([-1,num_channels])
#   channel_min = np.zeros(num_channels)
#   channel_max = np.zeros(num_channels)
#   for i in range(num_channels):
#       if i == dapi_index:
#           channel_min[i] = 0
#           channel_max[i] = I[:,i].max()
#       else:
#           channel_min[i] = np.percentile(I[:,i][I[:,i] > eps], 100-ptile)
#           channel_max[i] = np.percentile(I[:,i][I[:,i] > eps], ptile)
#   print(channel_min)
#   print(channel_max)
#   Images = (Images - channel_min)/channel_max
#   Images[Images < 0] = 0
#   Images[Images > 1] = 1
#   return Images*max_value

def rescale_16bit(image):
    """
    """
    # TODO: Considering reworking to be robust for bright/dark spots
    ii16 = np.iinfo(np.uint16)

    float_precision_image = image.astype(np.float32)
    rescaled_image = (float_precision_image - float_precision_image.min()) / float_precision_image.max()
    #print(ii16.max)
    #print(rescaled_image.max())

    return (rescaled_image * ii16.max).astype(np.uint16)

def register_images_all_samples(config, parallelize=0):
    """
    """
    # TODO: write docstring

    workspace_directory = config["workspace_directory"]
    input_directory = Path(workspace_directory, "spots_only")
    output_directory = Path(workspace_directory, "registered")
    output_directory.mkdir(exist_ok=True)

    samples = config["samples"]

    # TODO: Figure out if it's possible to parallelize by sample here.
    if parallelize > 0:
        num_processes = mp.cpu_count()
        print(num_processes)
        processes = []
        for sample in samples:
            process = mp.Process(target=register_images_single_sample, args=(sample, input_directory, output_directory))
            process.start()
            processes.append(process)

        for process in processes:
            process.join()
    else:
        for sample in samples:
            register_images_single_sample(sample, input_directory, output_directory)

def register_images_single_sample(sample, input_directory, output_directory):
    """
    """
    # TODO: write docstring
    
    sample_name = sample["name"]

    sample_input_subdirectory = input_directory / sample_name
    sample_output_subdirectory = output_directory / sample_name
    sample_output_subdirectory.mkdir(exist_ok=True)

    def build_reference_image_stack(rounds):
        """
        Helper function.
        """

        reference_filepaths = []
        for round_index, imaging_round in enumerate(rounds, start=1):
            round_directory = ("round%d" % round_index)
            round_input_subdirectory = sample_input_subdirectory / round_directory
            # round_input_subdirectory.mkdir(exist_ok=True)

            channels = imaging_round["channels"]
            reference_channel_index = imaging_round["reference_channel"]

            reference_filepaths.append((round_input_subdirectory / channels[reference_channel_index]).with_suffix( ".tif"))

        reference_image_stack = [imageio.imread(reference_filepath) for reference_filepath in reference_filepaths]

        # Crop all reference images to same size
        cropped_shape = tuple(np.array([image.shape for image in reference_image_stack]).min(axis = 0))
        print(cropped_shape)
        shape_crop_index = (slice(0, cropped_shape[0]), slice(0, cropped_shape[1]))

        reference_image_stack = np.stack([reference_image[shape_crop_index] for reference_image in reference_image_stack])
        print(reference_image_stack.shape)

        return reference_image_stack, shape_crop_index

    rounds = sample["rounds"]
    reference_image_stack, shape_crop_index = build_reference_image_stack(rounds)

    first_reference, *other_references = reference_image_stack
    required_shifts = [(0, 0)] + [find_shift(first_reference, other_reference) for other_reference in other_references]
    shift_crop_index = get_shift_crop_index(required_shifts)

    for round_index, imaging_round in enumerate(rounds, start=1):
        shift_index = round_index - 1
        round_directory = ("round%d" % round_index)
        channels = imaging_round["channels"]
        round_input_subdirectory = sample_input_subdirectory / round_directory
        round_output_subdirectory = sample_output_subdirectory / round_directory
        round_output_subdirectory.mkdir(exist_ok=True)

        # round_stack = build_round_stack(imaging_round, sample_input_subdirectory)
        # shifted_round_stack = apply_shift_to_stack(round_stack, required_shifts[round_index])

        for channel in channels:
            channel_filename = channel + ".tif"
            input_path = round_input_subdirectory / channel_filename
            output_path = round_output_subdirectory / channel_filename

            channel_image = imageio.imread(input_path)[shape_crop_index]
            registered_image = apply_shift(channel_image, required_shifts[shift_index])
            cropped_registered_image = registered_image[shift_crop_index]

            print(shape_crop_index)
            print(shift_crop_index)
            print(registered_image.shape)
            print(cropped_registered_image.shape)

            # TODO: Originally, we rescaled the intensity. Do we want to keep that here?
            scaled_image = rescale_16bit(cropped_registered_image)

            imageio.imwrite(output_path, scaled_image)

# def register_images_single_sample(basepath, tile_size=256, rounds_to_channels, dapi_index=3, processed_filename='DAPI.tiff', tmppath=None, rescale_intensity=True, save_tiles=True, save_stitched=True, shift_blocks=True):
#     """
#     Register images across multiple rounds in a single tissue sample.
# 
#     """
# 
#     pass
# 
# if __name__ == '__main__':
#   parser = argparse.ArgumentParser()
#   parser.add_argument('--basepath', help='Path to directory with subdirs for parsed images in each tissue')
#   parser.add_argument('--tissues', help='Comma-separated list of tissue numbers to include')
#   parser.add_argument('--tile-size', help='Size of output images',type=int,default=256)
#   parser.add_argument('--rounds-to-channels', help='CSV list of tissue, round, composition labels')
#   parser.add_argument('--dapi-index', help='Channel index for DAPI images',type=int,default=3)
#   parser.add_argument('--processed-filename', help='Name of processed files, or eg DAPI.tiff if separated by channel',default='DAPI.tiff')
#   parser.add_argument('--tmppath', help='Tmp path to store arrays_aligned_filtered',default=None)
#   parser.add_argument('--rescale-intensity',help='Rescale pixel intensities according the (nonzero) 1st,99th ptile observed in each channel', dest='rescale_intensity', action='store_true')
#   parser.add_argument('--save-tiles', dest='save_tiles', action='store_true')
#   parser.add_argument('--save-stitched', dest='save_stitched', action='store_true')
#   parser.add_argument('--shift-blocks',help='Find shifts in blocks. Used when stitching has failed in some round.', dest='shift_blocks', action='store_true')
#   parser.set_defaults(rescale_intensity=False)
#   parser.set_defaults(save_tiles=False)
#   parser.set_defaults(save_stitched=False)
#   parser.set_defaults(shift_blocks=False)
#   args,_ = parser.parse_known_args()
# 
#   Rounds2Channels = parse_rounds(args.rounds_to_channels)
#   for tissue_index in args.tissues.split(','):
#       tissue = 'tissue%s' % tissue_index
#       dapi_filepaths = glob.glob(os.path.join('%s/%s/round*/%s' % (args.basepath,tissue,args.processed_filename)))
#       dapi_filepaths = sorted(dapi_filepaths)
# 
#       dapi_images = [imageio.imread(dapi_filepath) for dapi_filepath in dapi_filepaths]
# 
#       if 'DAPI' in args.processed_filename:
#           for index in range(len(dapi_filepaths)):
#               round = dapi_filepaths[index].split('/')[-2]
#               channels = Rounds2Channels[(tissue,round)]
#               im = [imageio.imread(dapi_filepaths[index].replace('DAPI', channel)) for channel in channels]
#               im.insert(args.dapi_index, dapi_images[index])
#               dapi_images[i] = np.transpose(im, axes=(1,2,0))
# 
#         # Cut out reference images to all be the same size (TODO: is this necessary?)
#       min_shape = (min(im.shape[0] for im in dapi_images), min(im.shape[1] for im in dapi_images))
#       dapi_images = [im[:min_shape[0], :min_shape[1]] for im in dapi_images]
# 
#       if args.shift_blocks:
#           ImagesAligned = find_and_apply_shift_to_stack(dapi_images[0], dapi_images[1:], args.dapi_index)
#       else:
#           Images1 = [im[:,:,args.dapi_index] for im in dapi_images]
#           shifts = find_all_shifts(Images1[0],Images1[1:])
#           ImagesAligned = []
#           for i,im in enumerate(dapi_images):
#               if i == 0:
#                   ImagesAligned.append(im.astype(np.float32))
#               else:
#                   ImagesAligned.append(apply_shift_to_stack(im,shifts[i-1]))
#           ImagesAligned = [crop_all_shifted(images,shifts) for images in ImagesAligned]
#       
#         # Rescaling all intensities to lie in the range of 0-1
#         max_value = np.iinfo(dapi_images[0].dtype).max
# 
#       if args.rescale_intensity:
#           ImagesAligned = rescale_intensities(ImagesAligned,max_value,args.dapi_index)
# 
#       if args.save_stitched:
#           _=os.system('mkdir %s/%s/stitched_aligned_filtered' % (args.basepath,tissue))
#           for images,filepath in zip(ImagesAligned,dapi_filepaths):
#               chan_idx = np.where(images.reshape([-1,images.shape[-1]]).sum(0) > 0)[0]
#               im = images[:,:,chan_idx]
#               round = filepath.split('/')[-2]
#               channels = Rounds2Channels[(tissue,round)]
#               if (round == 'round1') and ('DAPI' not in channels):
#                   channels.insert(args.dapi_index,'DAPI')
#               for i,c in enumerate(channels):
#                   imageio.imwrite('%s/%s/stitched_aligned_filtered/%s.tiff' % (args.basepath,tissue,c), np.rint(im[:,:,i]).astype(dapi_images[0].dtype))
# 
#       if args.save_tiles:
#           if args.tmppath is None:
#               array_path = args.basepath
#           else:
#               array_path = args.tmppath
#               _=os.system('mkdir %s/%s' % (array_path,tissue))
#           _=os.system('mkdir %s/%s/arrays_aligned_filtered' % (array_path,tissue))
#           for images, filepath in zip(ImagesAligned,dapi_filepaths):
#               images = normalize_image_scale(images,max_value)
#               chan_idx = np.where(images.reshape([-1,images.shape[-1]]).sum(0) > 0)[0]
#               im = images[:, :, chan_idx]
#               round = filepath.split('/')[-2]
#            e  channels = Rounds2Channels[(tissue,round)]
#               if (round == 'round1') and ('DAPI' not in channels):
#                   channels.insert(args.dapi_index,'DAPI')
#               tiles, new_fov_pattern = break_into_tiles(im,args.tile_size)
#               new_fov_pattern = np.array(new_fov_pattern)
#               for tile, new_fov in zip(tiles,new_fov_pattern.flatten()):
#                   for i,c in enumerate(channels):
#                       #new_path = '%s/%s/arrays_aligned_filtered/fov_%d.%s.npy' % (args.basepath,tissue,new_fov,c)
#                       #np.save(new_path,tile[:,:,i])
#                       new_path = '%s/%s/arrays_aligned_filtered/fov_%d.%s.tiff' % (array_path,tissue,new_fov,c)
#                       imageio.imwrite(new_path,tile[:,:,i])
#               np.save('%s/%s/modified_fov_pattern.npy' % (args.basepath,tissue), new_fov_pattern)
#       print(tissue)



import numpy as np
from scipy.sparse import save_npz, load_npz, csr_matrix
from scipy.spatial import distance

CellMasks = load_npz('%s/%s/segmented/cell_masks.npz' % (args.mask_basepath,tissue))
FP_mask = glob.glob(os.path.join(args.mask_basepath,tissue,'segmented','DAPI.tiff'))
n = imageio.imread(FP_mask[0]).shape
CellCentroids = []
for cell in CellMasks:
	rows,cols = np.unravel_index(cell.indices,n)
	avg_r = np.average(rows)
	avg_c = np.average(cols)
	CellCentroids.append((avg_r,avg_c))

CellCentroids = np.array(CellCentroids)
CellDist = distance.squareform(distance.pdist(CellCentroids))
thresh = 80
CellNeighbors = csr_matrix((CellDist + np.eye(CellCentroids.shape[0])*1e9 < thresh))



import numpy as np
import imageio
from scipy.ndimage import gaussian_filter
from scipy.sparse import csr_matrix, save_npz, load_npz

import subprocess
from cellprofiler_core import image, object, pipeline, preferences, workspace, measurement
preferences.set_headless()

from pathlib import Path, PurePath
from importlib.resources import path
import multiprocessing as mp

def segment_single_sample(sample, input_directory, output_directory):
    """
    """
    # TODO: add docstring

    sample_name = sample["name"]
    sample_input_subdirectory = input_directory / sample_name
    sample_output_subdirectory = output_directory / sample_name
    
    rounds = sample["rounds"]

    composite_filepaths, direct_filepaths = split_filepaths(rounds, sample_input_subdirectory)

    merged_composite_filepath = sample_input_subdirectory / "merged_composites.tif"
    merge_composites(composite_filepaths, merged_composite_filepath)
   
    # TODO: fix this to be better if possible?
    # Assume there is at least one round of imaging to get a filepath to one reference image
    reference_round = 2
    initial_round = rounds[reference_round]
    reference_channel_index = initial_round["reference_channel"]
    round_input_subdirectory = sample_input_subdirectory / ("round%d" % (reference_round + 1))
    reference_channel_name = initial_round["channels"][reference_channel_index]
    reference_image_filepath = (round_input_subdirectory / reference_channel_name).with_suffix( ".tif")

    print(reference_image_filepath)

    segment(reference_image_filepath, merged_composite_filepath, sample_input_subdirectory, sample_output_subdirectory)
    aggregated_masks_filepath = sample_output_subdirectory / "composite_mask.tiff"
    sparse_masks_filepath = sample_output_subdirectory / "sparse_masks.npz"
    parse_and_save_cell_masks(aggregated_masks_filepath, sparse_masks_filepath)
   
    composite_measurements_filepath = sample_output_subdirectory / "composite_measurements.npy" 
    direct_measurements_filepath = sample_output_subdirectory / "direct_measurements.npy" 
    integrate_cells(sparse_masks_filepath, composite_filepaths, composite_measurements_filepath)
    integrate_cells(sparse_masks_filepath, direct_filepaths, direct_measurements_filepath)

def split_filepaths(rounds, input_directory):
    """
    """
    composite_filepaths = []
    direct_filepaths = []
    for round_index, imaging_round in enumerate(rounds, start=1):
        round_directory = ("round%d" % round_index)
        round_input_subdirectory = input_directory / round_directory

        channels = imaging_round["channels"]
        composite_channel_indices = imaging_round.get("composite_channels", [])
        reference_channel_index = imaging_round["reference_channel"]

        for channel_index, channel in enumerate(channels):
            if channel_index == reference_channel_index:
                continue
            elif channel_index in composite_channel_indices:
                composite_filepaths.append((round_input_subdirectory / channel).with_suffix( ".tif"))
            else:
                direct_filepaths.append((round_input_subdirectory / channel).with_suffix( ".tif"))

    composite_filepaths.sort(key=lambda filepath : filepath.name)
    direct_filepaths.sort(key=lambda filepath : filepath.name)

    return composite_filepaths, direct_filepaths

def segment_all_samples(config, parallelize=0):
    """
    """
    # TODO: write docstring

    workspace_directory = config["workspace_directory"]
    input_directory = Path(workspace_directory, "registered")
    output_directory = Path(workspace_directory, "segmented")
    output_directory.mkdir(exist_ok=True)

    samples = config["samples"]

    # TODO: Figure out if it's possible to parallelize by sample here.
    if parallelize > 0:
        num_processes = mp.cpu_count()
        processes = []
        for sample in samples:
            process = mp.Process(target=segment_single_sample, args=(sample, input_directory, output_directory))
            process.start()
            processes.append(process)

        for process in processes:
            process.join()
    else:
        for sample in samples:
            segment_single_sample(sample, input_directory, output_directory)

def merge_composites(composite_filepaths, output_path):
    """
    """

    first_filepath, *other_filepaths = composite_filepaths

    merged_image = imageio.imread(first_filepath)
    for composite_filepath in other_filepaths:
        other_image = imageio.imread(composite_filepath)
        merged_image = np.maximum(merged_image, other_image)

    filtered_merged_image = gaussian_filter(merged_image, 4)  
    imageio.imwrite(output_path, filtered_merged_image)

    return filtered_merged_image
   
def segment(reference_image_filepath, merged_composite_filepath, input_directory, output_directory):
    """
    """
    # TODO: this is hacky. Consult with CellProfiler people to use the Python interface to create
    # a cell segmentation pipeline
    if __package__:
        with path(__package__, "segment.cppipe") as segmentation_macro_filepath:
            pipeline_filepath = segmentation_macro_filepath
    else:
        pipeline_filepath = str(PurePath(__file__).parent / "segment.cppipe")

    subprocess.run(["cellprofiler", "-c", "-p", pipeline_filepath, "-i", input_directory, "-o", input_directory])
    subprocess.run(["mv", "output/composite_mask.tiff", output_directory])
    subprocess.run(["rm", "-rf", "output/"])
    
    # segmentation_pipeline = pipeline.Pipeline()
    # segmentation_pipeline.load("segment.cppipe")

    # image_set_list = image.ImageSetList()
    # image_set = image_set_list.get_image_set(0)

    # reference_image = imageio.imread(reference_image_filepath)
    # reference_handle = image.Image(reference_image)
    # merged_composite = imageio.imread(merged_composite_filepath)
    # merged_composite_handle = image.Image(merged_composite)

    # image_set.add("DAPI", reference_handle)
    # image_set.add("RNA", merged_composite_handle)
    # 
    # object_set = object.ObjectSet()

    # objects  = object.Objects()
    # 
    # object_set.add_objects(objects, "example")
    # measurements = measurement.Measurements()
    # 
    # segmentation_workspace = workspace.Workspace(
    #     segmentation_pipeline,
    #     None,
    #     image_set,
    #     object_set,
    #     measurements,
    #     image_set_list,
    # )
    # output_measurements = segmentation_pipeline.run(None)

def parse_and_save_cell_masks(input_path, output_path):
    """
    """
   
    aggregate_masks = imageio.imread(input_path)
    num_cells = np.unique(aggregate_masks).size
    flattened_masks = aggregate_masks.flatten()
    image_shape = flattened_masks.shape

    # TODO: this only works so far if there are a max of np.iinfo(np.uint16).max cell objects.
    # Amend so that it works for more as well???
    is_cell_index = (flattened_masks != 0)
    binary_mask_data = np.ones(shape=is_cell_index.sum(), dtype=np.float32)
    pixel_indices = is_cell_index.nonzero()
    mapping, cell_indices = np.unique(flattened_masks, return_inverse=True)
    #cell_indices = flattened_masks[pixel_indices] - 1

    compressed_cell_masks = csr_matrix((binary_mask_data, (cell_indices, *pixel_indices)), shape = (num_cells, *image_shape), dtype=np.float32)

    save_npz(output_path, compressed_cell_masks)

    return compressed_cell_masks

# def filter(im, filter_size):
#   if filter_size > 0:
#       if len(im.shape) == 3:
#           for i in range(im.shape[2]):
#               im[:,:,i] = ndimage.gaussian_filter(im[:,:,i], filter_size)
#       else:
#           im = ndimage.gaussian_filter(im, filter_size)
#   return im
# 
# if __name__ == '__main__':
#   parser = argparse.ArgumentParser()
#   parser.add_argument('--basepath', help='Path to directory with subdirs for parsed images in each tissue')
#   parser.add_argument('--tissues', help='Comma-separated list of tissue numbers to include')
#   parser.add_argument('--filter-size', help='Size of gaussian filter',type=float,default=4)
#   parser.add_argument('--stitched-subdir', help='Subdirectory with stitched images', default='stitched_aligned_filtered')
#   args,_ = parser.parse_known_args()
#   for t in args.tissues.split(','):
#       tissue = 'tissue%s' % t
#       FP = glob.glob(os.path.join(args.basepath,tissue,args.stitched_subdir,'Composite_*'))
#       im = imageio.imread(FP[0])
#       for fp in FP[1:]:
#           im = np.maximum(im,imageio.imread(fp))
#       im = filter(im, args.filter_size)
#       imageio.imwrite('%s/%s/%s/All_Composite.tiff' % (args.basepath,tissue,args.stitched_subdir),im)
#       print(tissue)
# if __name__ == '__main__':
#   parser = argparse.ArgumentParser()
#   parser.add_argument('--basepath', help='Path to directory with subdirs for parsed images in each tissue')
#   parser.add_argument('--tissues', help='Comma-separated list of tissue numbers to include')
#   parser.add_argument('--mask-file', help='Mask file (or prefix), eg ExpandedNuclei_* or DAPI.tiff', default='All_Composite.tiff')
#   args,_ = parser.parse_known_args()
#   for t in args.tissues.split(','):
#       tissue = 'tissue%s' % t
#       im = imageio.imread('%s/%s/segmented/%s' % (args.basepath,tissue,args.mask_file)).flatten()
#       col_ind = np.where(im > im.min())[0]
#       data = np.ones(len(col_ind),dtype=np.float32)
#       row_ind = im[col_ind].astype(np.int)-1
#       X = csr_matrix((data,(row_ind,col_ind)),shape=[len(np.unique(row_ind)),len(im)],dtype=np.float32)
#       save_npz('%s/%s/segmented/cell_masks.npz' % (args.basepath,tissue),X)
#       print(tissue,X.shape,len(data))

def integrate_cells(sparse_masks_filepath, image_filepaths, output_filepath):
    """
    """

    cell_masks = load_npz(sparse_masks_filepath)
    num_images = len(image_filepaths)
    num_cells = cell_masks.shape[0]
    measurements = np.zeros(shape=(num_images, num_cells), dtype=np.float32)
    for image_index, image_filepath in enumerate(image_filepaths):
        flattened_image = imageio.imread(image_filepath).flatten()
        integrated_intensities = cell_masks.dot(flattened_image)
        measurements[image_index] = integrated_intensities

    np.save(output_filepath, measurements)

    return measurements


# if __name__ == '__main__':
#     parser = argparse.ArgumentParser()
#     parser.add_argument('--basepath', help='Path to directory with subdirs for parsed images in each tissue')
#     parser.add_argument('--tissues', help='Comma-separated list of tissue numbers to include')
#     parser.add_argument('--stitched-subdir', help='Subdirectory with stitched images', default='stitched_aligned_filtered')
#     parser.add_argument('--mask-basepath', help='Optional alternative basepath for cell masks', default=None)
#     parser.add_argument('--as-average', dest='as_average', action='store_true')
#     parser.add_argument('--binary', dest='binary', action='store_true')
#     parser.set_defaults(as_average=False)
#     parser.set_defaults(binary=False)
#     args,_ = parser.parse_known_args()
# 
#     CellMasks = load_npz('%s/%s/segmented/cell_masks.npz' % (args.mask_basepath,tissue))
#     # if args.as_average:
#     #     for i in range(CellMasks.shape[0]):
#     #         ii = CellMasks.indptr[i]
#     #         ij = CellMasks.indptr[i+1]
#     #         CellMasks.data[ii:ij] = 1/(ij-ii)
# 
#     FP = glob.glob(os.path.join(args.basepath,tissue,args.stitched_subdir,'*.tiff'))
#     FP_composites = [fp for fp in FP if 'Composite_' in fp.split('/')[-1]]
#     FP_direct = [fp for fp in FP if (('Composite_' not in fp.split('/')[-1]) and ('All_Composite' not in fp.split('/')[-1]) and ('DAPI' not in fp.split('/')[-1]) and ('merged' not in fp.split('/')[-1]))]
#     # sort the composites by number
#     FP_composites = [(int(fp.split('_')[-1].split('.')[0]),fp) for fp in FP_composites]
#     FP_composites = [fp[1] for fp in sorted(FP_composites)]
#     X = []
#     L = []
#     for fp in FP_composites:
#         im = imageio.imread(fp).flatten()
#         if args.binary:
#             im = (im > 0).astype(im.dtype)
#         X.append(CellMasks.dot(im))
#         L.append(fp.split('/')[-1].split('.')[0])
#     # if args.as_average:
#     #     if args.binary:
#     #         savepath = '%s/%s/average_intensity_binary' % (args.basepath,tissue)
#     #     else:
#     #         savepath = '%s/%s/average_intensity' % (args.basepath,tissue)
#     # else:
#     #     if args.binary:
#     #         savepath = '%s/%s/integrated_intensity_binary' % (args.basepath,tissue)
#     #     else:
#     #         savepath = '%s/%s/integrated_intensity' % (args.basepath,tissue)
#     savepath = '%s/%s/integrated_intensity' % (args.basepath,tissue)
#     _=os.system('mkdir %s' % savepath)
#     np.save('%s/composite_measurements.npy' % savepath,X)
#     # np.save('%s/composite_labels.npy' % savepath,L)
#     X = []
#     L = []
#     # get the size of cell masks
#     FP_mask = glob.glob(os.path.join(args.mask_basepath,tissue,'segmented','ExpandedNuclei_*'))
#     if len(FP_mask) == 0:
#         FP_mask = glob.glob(os.path.join(args.mask_basepath,tissue,'segmented','All_Composite.tiff'))
#     n = imageio.imread(FP_mask[0]).shape
#     for fp in FP_direct:
#         im = imageio.imread(fp)
#         # images might have been padded on the bottom and right
#         im = im[:n[0],:n[1]].flatten()
#         if args.binary:
#             im = (im > 0).astype(im.dtype)
#         X.append(CellMasks.dot(im))
#         L.append('.'.join(fp.split('/')[-1].split('.')[:-1]))
#     if (tissue == 'tissue2') and ('Foxp2' in L): # Foxp2 was measured twice in tissue2
#         gi = L.index('Foxp2')
#         X.append(X[gi])
#         L.append(L[gi])
#     if args.as_average:
#         np.save('%s/direct_measurements.npy' % savepath,X)
#         np.save('%s/direct_labels.npy' % savepath,L)
#     else:
#         np.save('%s/direct_measurements.npy' % savepath,X)
#         np.save('%s/direct_labels.npy' % savepath,L)
# 
# 
# 
# 



from util import *

def stitch_from_coords(tissue_directory_regex, height, width, ordered_channels, filter_size = 8):

    round_subdirectory_string = "round*/"
    registered_coordinates_filename = "stitching_coordinates.registered.txt"
    round_directory_regex = os.path.join(tissue_directory_regex, round_subdirectory_string)

    MAX_INTENSITY = np.iinfo(np.uint16).max

    round_directories = glob.glob(round_directory_regex)

    max_coords = []

    for round_directory in round_directories:
        composite_directories = glob.glob(os.path.join(round_directory, "*", ""))
        num_channels = len(composite_directories)
        stitched_coordinates_path = os.path.join(round_directory, registered_coordinates_filename)
        _, max_x, max_y = parse_coordinates_file(stitched_coordinates_path)
        max_coords.append((max_x + width, max_y + height))

    min_global_width, min_global_height = np.amin(max_coords, axis = 0)

    for round_directory in round_directories[40:]:
        composite_directories = glob.glob(os.path.join(round_directory, "*", ""))
        num_channels = len(ordered_channels)
        stitched_coordinates_path = os.path.join(round_directory, registered_coordinates_filename)
        filename_to_coordinates, max_x, max_y = parse_coordinates_file(stitched_coordinates_path)
        
        canvas = np.zeros((max_y + height, max_x + width, num_channels), np.float64)
        #for composite_index, composite_directory in enumerate(composite_directories):
        #    print(composite_directory)
        #    channel_label, channel = os.path.basename(os.path.normpath(composite_directory)).split('.')

        for channel_index, channel in enumerate(ordered_channels):
            composite_directory_regex = round_directory + "*" + channel + "*"
            # composite_directory = glob.glob(composite_directory_regex)
            composite_directory = next(glob.iglob(composite_directory_regex))

            for filename in filename_to_coordinates:
                current_x, current_y = filename_to_coordinates[filename]
                source_image = imageio.imread(os.path.join(composite_directory, filename))
                canvas[current_y : current_y + height, current_x : current_x + width, channel_index] = source_image 
            
            
        reduced_canvas = normalize_and_convert_to_16_bit(canvas)
        filtered_reduced_canvas = median_filter(reduced_canvas, (filter_size, filter_size, 1)) 
        #print(np.sum(np.sum(filtered_reduced_canvas, axis=0), axis = 0))
        #print(filtered_reduced_canvas.shape)
        filtered_reduced_cropped_canvas = filtered_reduced_canvas[:min_global_height, :min_global_width]
        #print(filtered_reduced_cropped_canvas.shape)
        #print(filtered_reduced_cropped_canvas.shape)
        #print(filtered_reduced_canvas.ptp())
        imageio.imwrite(os.path.join(round_directory, "stitched.tiff"), filtered_reduced_cropped_canvas)

if __name__=="__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--tissue-directory-regex', help='Regex that matches all directories of composite images that must be stitched')
    parser.add_argument('--height', help='Height of eah FOV in pixels', type=int)
    parser.add_argument('--width', help='Width of each FOV in pixels', type=int)
    parser.add_argument('--ordered-channels', help='Channels, in order from lowest to highest frequency.', nargs='+')
    parser.add_argument('--filter-size', help='Dimensions of median filter', type=int, default=8)
    args, _ = parser.parse_known_args()
   
    stitch_from_coords(args.tissue_directory_regex, args.height, args.width, args.ordered_channels, args.filter_size) 



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cisipy", # Replace with your own username
    version="0.0.9",
    author="Shahul Alam",
    author_email="alam.shahul@gmail.com",
    description="Compressed imaging transcriptomics in Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alam-shahul/cisipy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "toml",
        "numpy",
        "scipy",
        "scikit-image",
        "tifffile",
        "nd2reader",
        "python-bioformats",
        "pyimagej",
        "starfish",
        "sympy<=1.5.1",
        #"spams"
    ],
    package_data={
        "preprocessing": ["Fuse.ijm", "segment.cppipe"],
        "": ["*.cppipe"]
    },
    python_requires='>=3.7',
)



import scyjava_config
scyjava_config.add_options('-Xmx8g')

import imagej
import multiprocessing as mp
import os
from pathlib import Path

from importlib.resources import open_text

if __package__:
    with open_text(__package__, "Fuse.ijm") as macro:
        STITCHING_MACRO = macro.read()
else:
    with open("Fuse.ijm") as macro:
        STITCHING_MACRO = macro.read() 

def stitch(input_directory, input_filename, output_directory, imagej_instance):
    """
    Use the BigStitcher plugin to stitch together the series at input_filepath.

    See Fuse.ijm for the ImageJ macro code.

    Args:
      input_directory (Union[str, pathlib.Path]):
        Directory in which input microscopy file is located.
      input_filename (str):
        Name of microscopy file (excluding file extension).
      output_directory (Union[str, pathlib.Path]):
        Directory in which to output stitched images,
      imagej_instance (imagej.ImageJ):
        An ImageJ instance.

    Returns:
      A handle on the Java process that is completing the stitching job. This is useful
      for keeping track of concurrent stitching jobs.
        
    """

    # Formatting arguments correctly for ImageJ macro code
    output_directory = str(output_directory)
    input_directory = f'{input_directory}/'

    args = imagej_instance.py.to_java({
      'basePath': input_directory,
      'filename': input_filename,
      'fusePath': output_directory
    })

    stitching_job = imagej_instance.script().run("macro.ijm", STITCHING_MACRO, True, args)

    return stitching_job

def stitch_all_samples(config, parallelize=0):
    """
    """
    # TODO: Fill in docstring

    workspace_directory = config["workspace_directory"]
    input_directory = Path(workspace_directory, "unstitched")
    output_directory = Path(workspace_directory, "stitched")
    output_directory.mkdir(exist_ok=True)

    samples = config["samples"]
    path_to_fiji = config["path_to_fiji"]

    # TODO: Figure out if it's possible to parallelize by sample here.
    
    imagej_instance = imagej.init(path_to_fiji)

    if parallelize > 1:
        num_processes = mp.cpu_count()
        processes = []
        for sample in samples:
            process = mp.Process(target=stitch_single_sample, args=(sample, input_directory, output_directory, imagej_instance, parallelize - 1))
            process.start()
            processes.append(process)
    
        for process in processes:
            process.join()
    else:
        for sample in samples:
            stitch_single_sample(sample, input_directory, output_directory, imagej_instance, parallelize)

def stitch_single_sample(sample, input_directory, output_directory, imagej_instance, parallelize=0):
    rounds = sample["rounds"]
    sample_name = sample["name"]

    sample_input_directory = input_directory / sample_name
    sample_output_directory = output_directory / sample_name
    sample_output_directory.mkdir(exist_ok=True)

    jobs = []
    for round_index, imaging_round in enumerate(rounds, start=1):
        round_directory = ("round%d" % round_index)
        filename = imaging_round["filename"]

        input_filename = (filename)
        output_directory = sample_output_directory / round_directory
        output_directory.mkdir(exist_ok=True)
        
        job = stitch(sample_input_directory, input_filename, output_directory, imagej_instance)
        if parallelize > 0:
            jobs.append(job)
        else:
            job.get()

    for job in jobs:
        job.get()



import numpy as np
import os
import argparse
import glob
from scipy.ndimage import gaussian_filter
from ast import literal_eval
import imageio

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--basepath', help='Path to directory with subdirs for parsed images in each tissue')
	parser.add_argument('--blank-round', help='Round in which blank images were collected (skip these)')
	parser.add_argument('--size', help='Pixels along each dimension of one fov',type=int,default=2048)
	parser.add_argument('--filter-width', help='Width of flat field gaussian_filter; eg 16 -> 1/16th of image size',type=int,default=8)
	parser.add_argument('--existing-flat-field', help='Path to existing flat field',default=None)
	args,_ = parser.parse_known_args()
	FP_blanks = glob.glob(os.path.join(args.basepath,'*',args.blank_round,'DAPI.conf-405','TileConfiguration.registered.txt'))
	FP = glob.glob(os.path.join(args.basepath,'*','*','DAPI.conf-405','TileConfiguration.registered.txt'))
	FP = [fp for fp in FP if fp not in FP_blanks]
	if args.existing_flat_field is None:
		flat_field = np.zeros((args.size,args.size))
		fov_count = 0
		for fp in FP:
			fovs = []
			coords = []
			f = open(fp)
			header = [f.readline() for _ in range(4)]
			for line in f:
				ls = line.strip().split(';')
				fovs.append(ls[0])
				coords.append(literal_eval(ls[2].strip()))
			f.close()
			coords = np.rint(coords).astype(np.int)
			coords[:,0] -= coords[:,0].min()
			coords[:,1] -= coords[:,1].min()
			im = imageio.imread(fp.replace('DAPI.conf-405/TileConfiguration.registered.txt','stitched.threshold.tiff'))
			im = np.average(im[:,:,:3],axis=2)
			for coord in coords:
				coord[0] = min(im.shape[1]-args.size,coord[0])
				coord[1] = min(im.shape[0]-args.size,coord[1])
				flat_field += im[coord[1]:coord[1]+args.size, coord[0]:coord[0]+args.size]
				fov_count += 1
		flat_field /= fov_count
		flat_field = gaussian_filter(flat_field, args.size//args.filter_width)
		print('done calculating flat field')
		np.save('%s/flat_field.npy' % args.basepath,flat_field)
	else:
		flat_field = np.load(args.existing_flat_field)
	inverse_field = 1/flat_field
	inverse_field /= inverse_field.max()
	for fp in FP:
		fovs = []
		coords = []
		f = open(fp)
		header = [f.readline() for _ in range(4)]
		for line in f:
			ls = line.strip().split(';')
			fovs.append(ls[0])
			coords.append(literal_eval(ls[2].strip()))
		f.close()
		coords = np.rint(coords).astype(np.int)
		coords[:,0] -= coords[:,0].min()
		coords[:,1] -= coords[:,1].min()
		im = imageio.imread(fp.replace('DAPI.conf-405/TileConfiguration.registered.txt','stitched.threshold.tiff'))
		im_corrected = np.zeros(im.shape)
		for coord in coords:
			coord[0] = min(im.shape[1]-args.size,coord[0])
			coord[1] = min(im.shape[0]-args.size,coord[1])
			x = np.transpose(im[coord[1]:coord[1]+args.size, coord[0]:coord[0]+args.size],axes=[2,0,1])*inverse_field
			im_corrected[coord[1]:coord[1]+args.size, coord[0]:coord[0]+args.size] = np.transpose(x,axes=[1,2,0])
		dt = im.dtype
		max_scale = np.iinfo(dt).max
		#for i in range(im_corrected.shape[2]):
		#	im_corrected[:,:,i] = np.rint(im_corrected[:,:,i]/im_max*max_scale)
		im_corrected = np.rint(im_corrected).astype(dt)
		outpath = fp.replace('DAPI.conf-405/TileConfiguration.registered.txt','stitched.threshold.corrected.tiff')
		imageio.imwrite(outpath,im_corrected)
		print(outpath)





import numpy as np
import spams
import pandas as pd
import scipy.spatial as sp, scipy.cluster.hierarchy as hc
import matplotlib
from scipy.spatial.distance import squareform

from pathlib import Path

matplotlib.use('Agg')
import matplotlib.pyplot as plt
matplotlib.style.use('ggplot')
#import seaborn as sns
from scipy.spatial import distance
from scipy.stats import entropy

def select_and_correct_comeasured(x,xc,phi,phi_corr,training_corr,phi_thresh=0.6,train_thresh=0.1):
    # find comeasured genes that are not coexpressed
    comeasured = []
    for i in range(phi_corr.shape[0]):
        xs = np.argsort(-phi_corr[i])
        for j in xs:
            if phi_corr[i,j] < phi_thresh:
                break
            comeasured.append((phi_corr[i,j],i,j))
    corrected_pairs = []
    for c in sorted(comeasured,reverse=True):
        if training_corr[c[1],c[2]] < train_thresh:
            x, both_nz = correct_coexpression(x,xc,phi,c[1],c[2])
            corrected_pairs.append((c[1],c[2], both_nz))
    return x, corrected_pairs

def correct_coexpression(x,xc,phi,i,j):
    # pick the gene with nearest expression pattern in scRNA
    thresh_i = np.percentile(x[i],99.9)/100
    thresh_j = np.percentile(x[j],99.9)/100
    both_nz = (x[i] > thresh_i)*(x[j] > thresh_j)
    dist = distance.cdist([phi[:,i], phi[:,j]], xc[:,both_nz].T,'correlation')
    i_closer = np.where(both_nz)[0][dist[0] < dist[1]]
    j_closer = np.where(both_nz)[0][dist[0] > dist[1]]
    x[i, j_closer] = 0
    x[j, i_closer] = 0
    return x, both_nz

def decompress_measurements_all_samples(config, parallelize=0):
    """
    """
    # TODO: write docstring

    workspace_directory = config["workspace_directory"]
    training_directory = Path(config["training_directory"])
    input_directory = Path(workspace_directory, "segmented")
    output_directory = Path(workspace_directory, "decompressed")
    output_directory.mkdir(exist_ok=True)

    samples = config["samples"]

    # TODO: Figure out if it's possible to parallelize by sample here.
    if parallelize > 0:
        num_processes = mp.cpu_count()
        processes = []
        for sample in samples:
            process = mp.Process(target=decompress_measurements_single_sample, args=(sample, input_directory, output_directory, training_directory))
            process.start()
            processes.append(process)

        for process in processes:
            process.join()
    else:
        for sample in samples:
            decompress_measurements_single_sample(sample, input_directory, output_directory, training_directory)

def decompress_measurements_single_sample(sample, input_directory, output_directory, training_directory):
    """
    """

    sample_name = sample["name"]
    sample_input_subdirectory = input_directory / sample_name
    sample_output_subdirectory = output_directory / sample_name
    sample_output_subdirectory.mkdir(exist_ok=True)

    phi = np.load(training_directory / "phi.npy")    
    num_measurements, num_genes = phi.shape
    
    gene_correlations = squareform(np.load(training_directory / "correlations.npy"))   
    gene_modules = np.load(training_directory / "gene_modules.npy")

    num_measurements_per_gene = phi.sum(axis=0)
    # TODO: question about this: why do we subtract the np.eye after and not before normalizing?
    phi_correlations = ((phi.T @ phi)/num_measurements_per_gene).T - np.eye(num_genes)

    composite_measurements_filepath = sample_input_subdirectory / "composite_measurements.npy"
    composite_measurements = np.load(composite_measurements_filepath)

    analysis_operator = phi.dot(gene_modules).astype(np.float32)
    synthesis_coefficients = sparse_decode(composite_measurements, analysis_operator, lasso_scaling_factor = 0.1, method='lasso')

    decompressed_measurements = gene_modules.dot(synthesis_coefficients)
    # TODO: below we have some "correction" due to the nature of our genomic data. Can we get rid of these edits?
    decompressed_measurements[np.isnan(decompressed_measurements)] = 0
    decompressed_measurements[decompressed_measurements < 0] = 0

    # TODO: figure out what corrected measurements are
    # if args.correct_comeasured:
    #   np.save('%s/ReconstructedImages/%s/%s/segmented.segmentation.npy' % (args.basepath, tissue, args.method), xhat)
    #   xhat,cp = select_and_correct_comeasured(xhat,composite_measurements,phi,phi_corr,train_corr)
    #   np.save('%s/ReconstructedImages/%s/%s/segmented.segmentation.adjusted.npy' % (args.basepath, tissue, args.method), xhat)
    # X1.append(xhat)

    print(sample_output_subdirectory / "decompressed_measurements.npy")
    
    np.save(sample_output_subdirectory / "decompressed_measurements.npy", decompressed_measurements)


def sparse_decode(analysis_coefficients, analysis_operator, lasso_scaling_factor=0.1, ridge_sparsity=None, minimum_loss=1., minimum_ridge_sparsity=0, method='omp'):
    """
    """
    print(analysis_coefficients.shape)
    print(analysis_operator.shape)
    analysis_coefficients_fortran = np.asfortranarray(analysis_coefficients)
    analysis_operator_fortran = np.asfortranarray(analysis_operator)
    coefficient_norm = np.linalg.norm(analysis_coefficients)

    num_bases, dimensionality = analysis_coefficients.shape
    _, latent_dimensionality = analysis_operator.shape
    if method == 'omp':
        if not ridge_sparsity:
            ridge_sparsity = min(num_bases, latent_dimensionality)  
        while ridge_sparsity < minimum_ridge_sparsity:
            sparse_synthesis_coefficients = spams.omp(analysis_coefficients_fortran, analysis_operator_fortran,
                          L=ridge_sparsity, numThreads=4)
            synthesis_coefficients = np.asarray(sparse_synthesis_coefficients.todense())
            loss = 1 - np.linalg.norm(analysis_coefficients - analysis_operator.dot(W))**2/coefficient_norm**2
            if loss < minimium_loss:
                break
            k -= 1
    elif method == 'lasso':
        print("Using lasso")
        lasso_penalty = (lasso_scaling_factor * coefficient_norm**2 / dimensionality)
        print("Lasso penalty is ", lasso_penalty)
        # TODO: set numThreads to some reasonable amount based on mp.cpu_count() and parallelization level
        sparse_synthesis_coefficients = spams.lasso(analysis_coefficients_fortran, analysis_operator_fortran,
                        lambda1=lasso_penalty, mode=1, numThreads=4, pos=False)
        print(sparse_synthesis_coefficients.shape)
        synthesis_coefficients = np.asarray(sparse_synthesis_coefficients.todense())
        print(synthesis_coefficients.shape)

    print(synthesis_coefficients.sum())

    return synthesis_coefficients

def get_sorted_direct_genes(rounds, input_directory):
    """
    """
    direct_genes = []
    for round_index, imaging_round in enumerate(rounds, start=1):
        round_directory = ("round%d" % round_index)
        round_input_subdirectory = input_directory / round_directory

        channels = imaging_round["channels"]
        composite_channel_indices = imaging_round.get("composite_channels", [])

        for channel_index, channel in enumerate(channels):
            if channel_index not in composite_channel_indices:
                direct_genes.append(channel)

    direct_genes.sort()

    return direct_genes

#if __name__ == '__main__':
#    AllGenes = np.load('%s/labels.genes.npy' % args.trainpath)
#    tissue = 'tissue%s' % t
#    composite_measurements = np.load('%s/FilteredCompositeImages/%s/%s/composite_measurements.npy' % (args.basepath, tissue,args.method))
#    direct_measurements = np.load('%s/FilteredCompositeImages/%s/%s/direct_measurements.npy' % (args.basepath, tissue,args.method))
#    direct_labels = np.load('%s/FilteredCompositeImages/%s/%s/direct_labels.npy' % (args.basepath, tissue, args.method))
#    W = sparse_decode(composite_measurements,phi.dot(U).astype(np.float32),0.1,method='lasso')
#    xhat = U.dot(W)
#    xhat[np.isnan(xhat)] = 0
#    xhat[xhat < 0] = 0
#
#    idx = [np.where(AllGenes == l)[0][0] for l in direct_labels]
#    xhat = xhat[idx]
#    n = direct_measurements.shape[0]+1
#    fig, axes = plt.subplots(max(2,int(np.floor(np.sqrt(n)))), int(np.ceil(np.sqrt(n))))
#    axes = axes.flatten()
#    plt.rcParams["axes.labelsize"] = 3
#    for i in range(n-1):
#        corr = 1-distance.correlation(direct_measurements[i],xhat[i])
#        sns_plt = sns.scatterplot(direct_measurements[i],xhat[i],ax=axes[i])
#        _= sns_plt.set(xlabel='Direct Intensity (arb. units)',ylabel='Recovered Intensity (arb. units)',title='%s; r=%.4f' % (direct_labels[i],corr))
#        print(tissue,direct_labels[i],corr)
#        C.append(corr)
#    corr = 1-distance.correlation(direct_measurements.flatten(),xhat.flatten())
#    sns_plt = sns.scatterplot(direct_measurements.flatten(),xhat.flatten(),ax=axes[-1])
#    _= sns_plt.set(xlabel='Direct Intensity (arb. units)',ylabel='Recovered Intensity (arb. units)',title='%s; r=%.4f' % ('all points',corr))
#    print(tissue,'all points',corr)
#
#    # _=plt.tight_layout()
#    # if args.correct_comeasured:
#    #   outpath = '%s/ReconstructedImages/%s/%s/scatter.segmented.heuristic_correction.png' % (args.basepath, tissue, args.method)
#    # else:
#    #   outpath = '%s/ReconstructedImages/%s/%s/scatter.segmented.png' % (args.basepath, tissue, args.method)
#    # fig.savefig(outpath)
#    # plt.close()
#
#    X0.append(direct_measurements.flatten())
#    X2.append(xhat.flatten())
#
#    X0 = np.hstack(X0)
#    X1 = np.hstack(X1)
#    X2 = np.hstack(X2)
#    corr = 1-distance.correlation(X0,X2)
#    sns_plt = sns.scatterplot(X0,X2)
#    _= sns_plt.set(xlabel='Direct Intensity (arb. units)',ylabel='Recovered Intensity (arb. units)',title='%s; r=%.4f' % ('all tissues, all points',corr))
#    print('all tissues, all points',corr)
#    print('average gene corr', np.average(C))
#    fig = sns_plt.get_figure()
#    if args.correct_comeasured:
#        outpath = '%s/ReconstructedImages/scatter.segmented.heuristic_correction.png' % (args.basepath)
#    else:
#        outpath = '%s/ReconstructedImages/scatter.segmented.png' % (args.basepath)
#    fig.savefig(outpath)
#    plt.close()
#    # gene-gene correlation
#    Corr = 1-distance.squareform(distance.pdist(X1,'correlation')) - np.eye(X1.shape[0])
#    df = pd.DataFrame(Corr,columns=AllGenes,index=AllGenes)
#    linkage = hc.linkage(distance.squareform(1-Corr-np.eye(X1.shape[0])), method='average')
#    sns_plt = sns.clustermap(df,mask=(df == 0), row_linkage=linkage, col_linkage=linkage)
#    sns_plt.savefig('%s/ReconstructedImages/gene_similarity.segmented.png' % args.basepath)
#    plt.close()
#    avg_ng = np.average([np.exp(entropy(x)) for x in X1.T if x.sum()>0])
#    print('average genes / cell:',avg_ng)
#
#
#    



import numpy as np
import os
import argparse
import glob
import imageio
from scipy.sparse import load_npz

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--basepath', help='Path to directory with subdirs for parsed images in each tissue')
	parser.add_argument('--tissues', help='Comma-separated list of tissue numbers to include')
	parser.add_argument('--stitched-subdir', help='Subdirectory with stitched images', default='stitched_aligned_filtered')
	parser.add_argument('--mask-basepath', help='Optional alternative basepath for cell masks', default=None)
	parser.add_argument('--as-average', dest='as_average', action='store_true')
	parser.add_argument('--binary', dest='binary', action='store_true')
	parser.set_defaults(as_average=False)
	parser.set_defaults(binary=False)
	args,_ = parser.parse_known_args()
	if args.mask_basepath is None:
		args.mask_basepath = args.basepath
	for t in args.tissues.split(','):
		tissue = 'tissue%s' % t
		CellMasks = load_npz('%s/%s/segmented/cell_masks.npz' % (args.mask_basepath,tissue))
		if args.as_average:
			for i in range(CellMasks.shape[0]):
				ii = CellMasks.indptr[i]
				ij = CellMasks.indptr[i+1]
				CellMasks.data[ii:ij] = 1/(ij-ii)
		FP = glob.glob(os.path.join(args.basepath,tissue,args.stitched_subdir,'*.tiff'))
		FP_composites = [fp for fp in FP if 'Composite_' in fp.split('/')[-1]]
		FP_direct = [fp for fp in FP if (('Composite_' not in fp.split('/')[-1]) and ('All_Composite' not in fp.split('/')[-1]) and ('DAPI' not in fp.split('/')[-1]) and ('merged' not in fp.split('/')[-1]))]
		# sort the composites by number
		FP_composites = [(int(fp.split('_')[-1].split('.')[0]),fp) for fp in FP_composites]
		FP_composites = [fp[1] for fp in sorted(FP_composites)]
		X = []
		L = []
		for fp in FP_composites:
			im = imageio.imread(fp).flatten()
			if args.binary:
				im = (im > 0).astype(im.dtype)
			X.append(CellMasks.dot(im))
			L.append(fp.split('/')[-1].split('.')[0])
		if args.as_average:
			if args.binary:
				savepath = '%s/%s/average_intensity_binary' % (args.basepath,tissue)
			else:
				savepath = '%s/%s/average_intensity' % (args.basepath,tissue)
		else:
			if args.binary:
				savepath = '%s/%s/integrated_intensity_binary' % (args.basepath,tissue)
			else:
				savepath = '%s/%s/integrated_intensity' % (args.basepath,tissue)
		_=os.system('mkdir %s' % savepath)
		np.save('%s/composite_measurements.npy' % savepath,X)
		np.save('%s/composite_labels.npy' % savepath,L)
		X = []
		L = []
		# get the size of cell masks
		FP_mask = glob.glob(os.path.join(args.mask_basepath,tissue,'segmented','ExpandedNuclei_*'))
		if len(FP_mask) == 0:
			FP_mask = glob.glob(os.path.join(args.mask_basepath,tissue,'segmented','All_Composite.tiff'))
		n = imageio.imread(FP_mask[0]).shape
		for fp in FP_direct:
			im = imageio.imread(fp)
			# images might have been padded on the bottom and right
			im = im[:n[0],:n[1]].flatten()
			if args.binary:
				im = (im > 0).astype(im.dtype)
			X.append(CellMasks.dot(im))
			L.append('.'.join(fp.split('/')[-1].split('.')[:-1]))
		if (tissue == 'tissue2') and ('Foxp2' in L): # Foxp2 was measured twice in tissue2
			gi = L.index('Foxp2')
			X.append(X[gi])
			L.append(L[gi])
		if args.as_average:
			np.save('%s/direct_measurements.npy' % savepath,X)
			np.save('%s/direct_labels.npy' % savepath,L)
		else:
			np.save('%s/direct_measurements.npy' % savepath,X)
			np.save('%s/direct_labels.npy' % savepath,L)
		print(tissue)







import numpy as np
import matplotlib as mpl
mpl.use('Agg')
from matplotlib import pyplot as plt
from matplotlib import cm
import imageio

f = open('../../Training/Data/37genes.txt')
Genes = [line.strip() for line in f]
f.close()

GeneGroups = {'Excitatory': ['Slc17a7','Rorb','Paqr8','Deptor','Foxp2','Tle4','Rgs12','Stard8','Cux2','Sulf1'],
				'Inhibitory': ['Gad1','Vip','Parm1','Sst','Grin3a','Pvalb','Ndnf','Fgf13'],
				'Astrocytes': ['Gja1','F3','Prex2'],
				'Microglia': ['Ctss','Csf1r','Hmha1'],
				'Oligodendrocytes': ['Olig1','Vcan','Olig2'],
				'OPCs': ['Pdgfra','Mog'],
				'Endothelial': ['Flt1','Xdh','Id1'],
				'SMC': ['Vtn','Colec12'],
				'Other glia': ['Pdgfd','Id3','Ly86']}

colors = cm.get_cmap('gist_ncar',len(Genes)+8)
Colors = np.zeros([len(Genes),3])
i = 6
GeneColors = {}
for k,v in GeneGroups.items():
	for g in v:
		Colors[Genes.index(g)] = colors(i)[:3]
		_=plt.scatter(10,(i+1)*50,c=colors(i),marker=r"$ {} $".format(g),edgecolors='none',s=100)
		i += 1

plt.savefig('gene_colors.pdf')
plt.close()
Colors = np.array(Colors)

for t in range(1,9):
	tissue = 'tissue%d' % t
	im_shape = imageio.imread('%s/stitched/%s.tiff' % (tissue,Genes[0])).shape
	Merged = np.zeros([im_shape[0],im_shape[1],3])
	X = []
	for gene in Genes:
		im = imageio.imread('%s/stitched/%s.tiff' % (tissue,gene))
		X.append(im)
	max_gene = np.argmax(X,axis=0)
	X_max = np.array(X).reshape([len(Genes),-1])[max_gene.flatten(),np.arange(np.product(X[0].shape))]
	Merged = (Colors[max_gene.flatten()].T*X_max).T.reshape([im_shape[0],im_shape[1],3])
	max_val = np.iinfo('uint16').max
	threshold = np.percentile(Merged[Merged > 0],99)
	Merged_scaled = np.rint(Merged/threshold*max_val).astype('uint16')
	imageio.imsave('%s/stitched/all_genes.merged.tiff' % tissue, Merged_scaled)
	print(tissue)





import json
from util import *
from parse_and_z_project import parse_and_z_project
from stitch_from_coords import stitch_from_coords
from imagej_stitching import imagej_stitching
from subtract_background import subtract_background
from apply_channel_thresholds import apply_channel_thresholds
from calculate_and_apply_flat_field_correction import calculate_and_apply_flat_field_correction
from register_and_tile_images import register_and_tile_images

"""
Steps taken during preprocessing:

0. run parse_and_z_project.py, which parses and max z-projects each nd2 file
1. use imagej_stitching.py to calculate stitching in DAPI channel for every tissue/round
2. use stitch_from_coords.py to stitch together images for each gene x round combination and apply median filter
3. run subtract_background
4. use Fiji to determine intensity thresholds for each channel based on previous stitched images (setting max to 150% of min)
5. run apply_channel_thresholds.py
6. run calculate_and_apply_flat_field_correction.py
7. run save_registered_images.py to register across rounds and break into tiles
8. run write_tf_records.py
"""

parser = argparse.ArgumentParser()
parser.add_argument('--parameter-filepath', help='Path to JSON file containing parameters necessary for scripts to run')
parser.add_argument('--start', help='Index of first script in pipeline to run', type=int)
parser.add_argument('--end', help='Index of last script in pipeline to run', type=int)
args, _ = parser.parse_known_args()

filepath = args.parameter_filepath
with open(filepath) as f:
    parameters = json.load(f)

parameter_specification = [
    ["background_path_regex", "source_path_regex", "fov_order_map_path", "context_to_channel_map_path", "outpath"],
    ["tissue_directory_regex", "cols", "rows"],
    ["tissue_directory_regex", "height", "width", "ordered_channels"],
    ["tissue_directory_regex", "ordered_channels", "background_scaling_factors", "blank_round_number"],
    ["tissue_directory_regex", "ordered_channels", "channel_threshold_filepath", "blank_round_number"],
    ["tissue_directory_regex", "blank_round_number", "height", "width", "filter_width"],
    ["tissue_directory_regex", "tile_size", "blank_round_number"]
]

function_specification = [
    parse_and_z_project,
    imagej_stitching,
    stitch_from_coords,
    subtract_background,
    apply_channel_thresholds,
    calculate_and_apply_flat_field_correction,
    register_and_tile_images
]

if args.start > args.end:
    raise ValueError("Last index must be larger than first index.")

for index in range(args.start, args.end + 1):
    if index > len(parameter_specification):
        raise ValueError("Not a valid script_index.")
        
    necessary_parameters = parameter_specification[index]
    missing_parameters = [key for key in necessary_parameters if key not in parameters]
    if missing_parameters:
        raise ValueError("The input JSON file is missing the following necessary parameters: {}".format(missing_parameters))
    
    script_function = function_specification[index]
    print("Starting script %d" % index)
    script_function(*[parameters[key] for key in necessary_parameters])
    print("Finished script %d!" % index)



import numpy as np
import imageio
import glob, os
import re
import argparse
import pickle
import json
import pyfftw

from scipy import ndimage
from scipy.ndimage import median_filter
from scipy.ndimage import fourier_shift
from scipy.ndimage import gaussian_filter
from scipy.misc import imsave
from skimage.feature import register_translation
from skimage.transform import rescale
from pims import ND2_Reader
from collections import defaultdict
from ast import literal_eval

def adjust_contrast(image, c0=10, c1=0.3, eps=1e-5, ptile=99.5):
    """
    Use a sigmoid model to adjust contrast for an image.
    """
    contrast_adjusted_image = image/(np.percentile(image, ptile) + eps)
    contrast_adjusted_image = 1/(1 + np.exp(-(contrast_adjusted_image - c1)*c0))
    return contrast_adjusted_image

def _bytes_feature(value):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

def _int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))

def write_record_from_numpy(image, tissue, fov, tfrecords_filename, patch_height=512, patch_width=512):
    tissue_byte = tf.compat.as_bytes(tissue)
    fov_byte = tf.compat.as_bytes(fov)
    writer = tf.python_io.TFRecordWriter(tfrecords_filename + '.tfrecords')
    for i in range(0,image.shape[0],patch_height):
        for j in range(0,image.shape[1],patch_width):
            image_raw = image[i:i+patch_height,j:j+patch_width]
            h,w,c = image_raw.shape
            image_raw = image_raw.tostring()
            example = tf.train.Example(features=tf.train.Features(feature={'raw_data': _bytes_feature(image_raw), 'height': _int64_feature(h), 'width': _int64_feature(w), 'channels': _int64_feature(c), 'tissue': _bytes_feature(tissue_byte), 'fov': _bytes_feature(fov_byte), 'row_offset': _int64_feature(i), 'col_offset': _int64_feature(j)}))
            writer.write(example.SerializeToString())
    writer.close()

def shift_background(im, im_bg):
    n,m = im.shape
    diff_min = np.inf
    for i in range(-20, 20, 2):
        for j in range(-20, 20, 2):
            im_bg_adj = im_bg[100+i:n-100+i, 100+j:m-100+j]
            diff = np.linalg.norm(im[100:n-100, 100:m-100] - im_bg_adj)
            if diff < diff_min:
                best = (i,j)
                diff_min = diff
    i,j = best
    im_bg_adj = im_bg[max(0,i):min(n,n+i),max(0,j):min(m,m+j)]
    im_bg_adj = np.pad(im_bg_adj,((max(0,-i),max(0,i)),(max(0,-j),max(0,j))),'constant')
    return im_bg_adj

# New code - 1/7/19
def find_random_background(im, n=2500, radius=16):
	# find a region with low total variation
	total_variations = []
	sub_ims = []
	rows = np.random.randint(radius, im.shape[0] - radius, size=n)
	cols = np.random.randint(radius, im.shape[1] - radius, size=n)
	for row, col in zip(rows,cols):
		im_sub = im[row-radius:row+radius, col-radius:col+radius]
        # This line does all the magic
		total_variation = np.average(np.diff(im_sub,axis=0)**2) + np.average(np.diff(im_sub,axis=1)**2)
		total_vartiations.append(total_variation)
		sub_ims.append(im_sub)
	xs = np.argsort(total_variations)
	return np.average([sub_ims[i] for i in xs[:10]])

# Newer code - 1/16/19
def parse_coordinates_file(filepath):
    filename_to_coordinates = {}
    max_x = max_y = 0

    with open(filepath, 'r') as descriptor:
        # Discard first four lines and last line
        coordinate_information = descriptor.read().split('\n')[4:-1]

        for line in coordinate_information:
            data = line.strip().split(';')
            print(data)
            fov = data[0]
            coordinate_pair = np.clip(np.rint(literal_eval(data[2].strip())).astype(int), 0, None)
 
            filename_to_coordinates[fov] = coordinate_pair

            x_coord, y_coord = coordinate_pair
            max_x = max(max_x, x_coord)
            max_y = max(max_y, y_coord)           

    return filename_to_coordinates, max_x, max_y


def find_shift(first_image, second_image, scale = 1):
    """
    Obtain shift that aligns second_image to first_image, after rescaling both images to given resolution
    """
    rescaled_first_image = rescale(first_image, scale)
    rescaled_second_image = rescale(second_image, scale)
    shift, error, phase_difference = register_translation(rescaled_first_image, rescaled_second_image)
    return shift / scale

def apply_shift_to_images(images, shift):
    """
    Apply translation to image. Applies the shift to the Fourier transform of the images,
    and then inverse Fourier transforms the result.
    """
    offset_images = [fourier_shift(np.fft.fftn(image), shift) for image in images]
    offset_images = [np.fft.ifftn(image) for image in offset_images]
    offset_images = np.array(offset_images).astype(np.float32)
    return offset_images

def crop_shifted_images(images, shift):
    """
    Crop images according to x/y shifts. 
    """

    x_shift, y_shift = shift

    # Calculate the minimum and maximum horizontal shift
    if x_shift < 0:
        images = images[:, :x_shift]
    elif x_shift > 0:
        images = images[:, x_shift:]
    if y_shift < 0:
        images = images[:, :, :y_shift]
    elif y_shift > 0:
        images = images[:, :, y_shift:]
    return images

def parse_nd2_file(nd2_filepath):
    """
    Parse images and metadata necessary for stitching from nd2 file.
    """

    with ND2_Reader(nd2_filepath) as images:
        #print("The metadata is ", str(images.metadata).encode('utf-8'))
        
        num_fov = images.sizes['m']
        num_channels = images.sizes['c']
        num_rows = images.sizes['y']
        num_columns = images.sizes['x']

        fields_of_view = list(range(num_fov))
        channels = [images.metadata['plane_' + str(num)]['name'] for num in range(num_channels)]
        microns_per_pixel = images.metadata['calibration_um']

        try:
            images.iter_axes = 'mc'
            images.bundle_axes = 'zyx'
        except:
            images.iter_axes = 'c'
            images.bundle_axes = 'zyx'

        aggregated_images = []
        coordinate_pairs = []
        for z_stack in images:
            aggregated_image = np.max(z_stack, axis = 0)
            aggregated_images.append(aggregated_image)
            coordinate_pair = z_stack.metadata['y_um'], z_stack.metadata['x_um']
            coordinate_pairs.append(coordinate_pair)

    aggregated_images = np.reshape(aggregated_images,(num_fov, num_channels, num_rows, num_columns))
    coordinate_pairs = np.average(np.reshape(coordinate_pairs, (num_fov, num_channels, 2)), axis = 1)
    print("Coordinate pairs is \n" + str(coordinate_pairs))
    print("Shape is " + str(aggregated_images.shape))
    
    data = {
               "aggregated_images": aggregated_images,
               "coordinate_pairs": coordinate_pairs,
               "fields_of_view": fields_of_view,
               "channels": channels,
               "microns_per_pixel": microns_per_pixel 
           }

    return data

def calculate_flat_field(images):
    """
    Calculates an approximation of the flat field for the microscope by obtaining a per-pixel median and
    then applying a Gaussian filter. 
    """

    _, height, _ = images.shape
    print(images.shape)
    filter_sigma = height // 16
    flat_field = gaussian_filter(normalize_and_convert_to_16_bit(np.median(images, axis = 0)), filter_sigma)
   
    print(flat_field.shape)
    return flat_field 

def normalize_and_convert_to_16_bit(image):
    """
    Normalize an image and convert it to a 16-bitdepth scale; usually used right before saving an image.
    """
    
    MAX_INTENSITY = np.iinfo(np.uint16).max
    reduced_image = (image.astype(np.float64) - image.min())/(image.ptp())
    reduced_image = MAX_INTENSITY * reduced_image
    reduced_image = reduced_image.astype(np.uint16)

    return reduced_image

# New code - 1/10/19
def parse_rounds(filepath):
	round_tissue_combination_to_channel_map = defaultdict(list)
	f = open(filepath,encoding='utf-8')
	header = f.readline()
	for line in f:
		tissue, round_number, channel_label = line.strip().split(',')
		round_tissue_combination_to_channel_map[(tissue, round_number)].append(channel_label)
	f.close()
	return round_tissue_combination_to_channel_map

def apply_shifts(image, shift):
    offset_image = []
    for channel_index in range(image.shape[2]):
        print(image[:, :, channel_index].shape)
        print(image[:, :, channel_index].dtype)
        print("Calculating Fourier transform...")
        #transformed_slice = np.fft.fftn(image[:, :, channel_index])
        transformed_slice = pyfftw.interfaces.numpy_fft.fftn(image[:, :, channel_index])
        print("Shifting transformed image...")
        offset_im = fourier_shift(transformed_slice, shift)
        print(offset_im.shape)
        print(offset_im.dtype)
        offset_image.append(pyfftw.interfaces.numpy_fft.ifftn(offset_im))
        #offset_image.append(np.fft.ifftn(offset_im))
    offset_image = np.transpose(offset_image, axes = [1, 2, 0]).astype(np.float32)
    return offset_image

def find_all_shifts(ref_image, other_images):
    shifts = []
    for other_image in other_images:
        shift = find_shift(ref_image, other_image, scale=0.25)
        shifts.append(shift)
    return shifts

def crop_all_shifted(images, shifts):
    min_y_shift, min_x_shift = np.min(shifts,axis=0).astype(np.int)
    max_y_shift, max_x_shift = np.max(shifts,axis=0).astype(np.int)
    if min_y_shift < 0:
        images = images[:, :min_y_shift]
    if max_y_shift > 0:
        images = images[:, max_y_shift:]
    if min_x_shift < 0:
        images = images[:, :, :min_x_shift]
    if max_x_shift > 0:
        images = images[:, :, max_x_shift:]
    return images

def tilify(image, tile_size):
    rows, columns, _ = image.shape
    tiles = []
    tile_pattern = []
    tile_index = 0
    for row in range(0, rows, tile_size):
        tile_row = []
        for column in range(0, columns, tile_size):
            cutout = image[row:row+tile_size, column:column+tile_size]
            actual_height, actual_width, _ = cutout.shape
            tile = np.pad(cutout, [(0, tile_size - actual_height), (0, tile_size - actual_width), (0, 0)], 'constant')
            tiles.append(tile)
            tile_row.append(tile_index)
            tile_index += 1
        tile_pattern.append(tile_row)
    tiles = np.array(tiles)
    tile_pattern = np.array(tile_pattern)    

    return tiles, tile_pattern

def apply_threshold_and_rescale(channel_slice, lower_threshold, upper_threshold):
    datatype = channel_slice.dtype
    max_value = np.iinfo(datatype).max

    channel_slice = np.clip(channel_slice, lower_threshold, upper_threshold)

    channel_slice = channel_slice - lower_threshold
    thresholded_slice = np.rint(channel_slice/channel_slice.ptp() * max_value).astype(datatype)

    return thresholded_slice

